
/* $Id: cpd_usr_mp.c,v 3.14 2000/06/12 21:53:52 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* File: cpd_usr_mp.c - message passing wrappers for users */


/* System Header Files */

#include <stdio.h>
#include <string.h>


/* STV Header Files */

#include "stv.h"
#include "stvmac.h"
#include "stv_glob.h"


/* ---------- stv_wrapRecv() -------- */
/* Old, poorly-named, goof-ball routine... */

int
stv_wrapRecv( src, op, tag )
int src;
int op;
int tag;
{
	return( stv_userRecv( src, op, tag ) );
}


/* ---------- stv_userRecv() -------- */

/*
   recv's a message. sets the active recv buffer to bufid, returns
   -100 if the Sender Failed;
   -1000 if it is a Restart;
*/

int
stv_userRecv( src, op, tag )
int src;
int op;
int tag;
{
	STV_CONTEXT ctxt;

	STV_BUFPTR msg;

	int crbuf;
	int flush;
	int gen;

	/* Clear Out User's Current Receive Buffer */
	/* (this routine replaces regular stv_recv() call) */

	stv_current_rbuf( &crbuf );

	if ( crbuf > 0 )
		stv_free_msg_buf( crbuf );

	stv_set_rbuf( 0 );

	/* Get the Current Generation */

	gen = stv_generation();

	/* Get the Current Message Context */

	stv_get_context( &ctxt );

	/* Check for Messages, Until Valid Generation or Failure */

	do
	{
		flush = 0;

		msg = stv_recvMessage( stvAny, op, src, tag, ctxt,
				stvMsgBlock );

		/* If NULL Message Returned, Something Went Awry or Not Found */

		if ( msg == (STV_BUFPTR) NULL )
		{
			STV_ERROR( "stv_userRecv", stvStatusSysFail,
					"Null Message Returned by stv_recvMessage()." );

			return( -10000 );
		}

		/* Single Sender Failed / Halted */

		if ( msg->op == stvNotify
				&& ( ( src != -1 && msg->src == src )
					|| msg->src == stv_cpdtid() ) )
		{
			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_MSG | STV_DEBUG_NOTIFY ) )
				printf( "stv_userRecv(): Sender / CPD Failed.\n" );

			/* Make Sure Notify Handlers Get Called */
			stv_vector_handler( STVHANDLERS, nSTVHANDLERS, msg );

			crbuf = -100;

			stv_free_msg_buf( msg->bufid );

			FREE( msg );
		}

		/* Old Generation, Flush It */

		else if ( msg->fgid < gen )
		{
			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_MSG | STV_DEBUG_FLUSH ) )
			{
				printf( "Flushing op %d with old gen=%d (gen=%d)\n",
						msg->op, msg->fgid, gen );
			}

			stv_free_msg_buf( msg->bufid );

			FREE( msg );

			flush++;
		}

		/* Handle Messages */

		else
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_MSG )
				printf( "stv_userRecv(): got msg op %s/%d\n",
						stv_msg_op_str( stvMsgOp( msg->op ) ),
						msg->op );

			if ( msg->op == stvRestart )
			{
				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_MSG | STV_DEBUG_NOTIFY ) )
					printf( "stv_userRecv(): Got Restart!\n" );

				crbuf = -1000;

				stv_free_msg_buf( msg->bufid );

				FREE( msg );
			}

			else
			{
				crbuf = msg->bufid;

				FREE( msg );
			}
		}
	}
	while ( flush );


	if ( crbuf > 0 )
		stv_set_rbuf( crbuf );

	return( crbuf );
}


/* ----- stv_userInitSend() ------*/

int
stv_userInitSend( type, op )
int type;
int op;
{
	int csbuf, gen;
	int tmp;

	stv_current_sbuf( &csbuf );

	if ( csbuf > 0 )
		stv_free_msg_buf( csbuf );

	if ( stv_init_msg_buf( type, &csbuf ) )
	{
		tmp = stvMsgOp( op );
		stv_pack_int( csbuf, &tmp, 1, 1 );

		gen = stv_generation();
		stv_pack_int( csbuf, &gen, 1, 1 );

		stv_set_sbuf( csbuf );
	}

	else
		stv_mp_err( "Creating User Message Buffer" );

	return( csbuf );
}


/* ----- stv_userSend() ------*/

int
stv_userSend( dest, tag )
int dest;
int tag;
{
	int csbuf;

	stv_current_sbuf( &csbuf );

	if ( csbuf > 0 )
		return( stv_send_msg( csbuf, dest, tag, STV_BASECTX ) );

	else
		return( stvStatusBadParam );
}


/* ----- stv_userMcast() ------*/

int
stv_userMcast( list, nlist, tag )
int *list;
int nlist;
int tag;
{
	int csbuf;

	stv_current_sbuf( &csbuf );

	if ( csbuf > 0 )
		return( stv_mcast_msg( csbuf, list, nlist, tag, STV_BASECTX ) );

	else
		return( stvStatusBadParam );
}

