
/* $Id: commands.c,v 3.5 2000/06/12 18:01:00 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


#include "flyme.h"
#include "flyme_glob.h"


/* ARGSUSED */
int set_play_mode_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: set_play_mode < Monitor | Play >",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	if ( !strcmp( argv[1], "Monitor" ) )
	{
		PLAY_MODE = MONITORING;

		release_heli();
	}

	else if ( !strcmp( argv[1], "Play" ) )
		PLAY_MODE = PLAYING;

	else
	{
		Tcl_SetResult( itp, "usage: set_play_mode < Monitor | Play >",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	set_display_scale( MAIN_WIN );

	adjust_border( MAIN_WIN );

	return( TCL_OK );
}


/* ARGSUSED */
int set_opts_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char *auto_steer;
	char *auto_zoom;

	char cmd[1024];

	if ( argc != 3 )
	{
		Tcl_SetResult( itp, "usage: set_opts <auto_steer> <auto_zoom>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	auto_steer = argv[1];
	auto_zoom = argv[2];

	if ( !strcmp( auto_steer, "ON" ) )
		AUTO_STEER = stvTrue;
	
	else
		AUTO_STEER = stvFalse;

	if ( !strcmp( auto_zoom, "ON" ) )
	{
		AUTO_ZOOM = stvTrue;

		sprintf( cmd, "%s.zoom_factor delete 0 end", FM );
		Tcl_Eval( itp, cmd );

		sprintf( cmd, "%s.zoom_factor insert 0 auto", FM );
		Tcl_Eval( itp, cmd );
	}
	
	else
	{
		AUTO_ZOOM = stvFalse;

		sprintf( cmd, "%s.zoom_factor delete 0 end", FM );
		Tcl_Eval( itp, cmd );

		sprintf( cmd, "%s.zoom_factor insert 0 %lg", FM, ZOOM_FACTOR );
		Tcl_Eval( itp, cmd );
	}

	return( TCL_OK );
}


/* ARGSUSED */
int set_zoom_factor_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char cmd[1024];

	double factor;

	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: set_zoom_factor <num>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	if ( AUTO_ZOOM )
	{
		if ( strcmp( argv[1], "auto" ) )
		{
			sprintf( cmd, "%s.zoom_factor delete 0 end", FM );
			Tcl_Eval( itp, cmd );

			sprintf( cmd, "%s.zoom_factor insert 0 auto", FM );
			Tcl_Eval( itp, cmd );
		}
	}

	else
	{
		if ( !strcmp( argv[1], "" ) )
		{
			sprintf( cmd, "%s.zoom_factor delete 0 end", FM );
			Tcl_Eval( itp, cmd );

			sprintf( cmd, "%s.zoom_factor insert 0 %lg",
				FM, ZOOM_FACTOR );

			Tcl_Eval( itp, cmd );
		}

		else if ( sscanf( argv[1], "%lf", &factor ) == 1 )
		{
			if ( !EQUIV( factor, ZOOM_FACTOR ) )
			{
				ZOOM_FACTOR = factor;

				if ( CURRENT_HELI != NULL )
					set_display_to_heli( CURRENT_HELI, stvTrue );
			}
		}

		else
		{
			Tcl_SetResult( itp, "usage: set_zoom_factor <num>",
				TCL_STATIC );

			return( TCL_ERROR );
		}
	}

	return( TCL_OK );
}


/* ARGSUSED */
int set_radar_range_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char cmd[1024];

	double range;

	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: set_radar_range <num>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	if ( sscanf( argv[1], "%lf", &range ) == 1 )
	{
		if ( range < 50.0 )
		{
			range = 50.0;

			sprintf( cmd, "%s.radar_range delete 0 end", FM );
			Tcl_Eval( itp, cmd );

			sprintf( cmd, "%s.radar_range insert 0 %lg", FM, range );
			Tcl_Eval( itp, cmd );
		}

		if ( !EQUIV( range, RADAR_RANGE ) )
		{
			RADAR_RANGE = range;

			if ( CURRENT_HELI != NULL )
			{
				if ( stv_viewer_steer_vparameter( CURRENT_VP,
					(STV_VALUE) &RADAR_RANGE, stvDouble,
					RADAR_RANGE_INDEX ) != stvStatusOk )
				{
					printf( "Error steering radar range...\n" );
				}
			}
		}
	}

	else
	{
		Tcl_SetResult( itp, "usage: set_radar_range <num>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	return( TCL_OK );
}


/* ARGSUSED */
int resize_stv_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	if ( CONNECTED )
	{
		set_display_scale( MAIN_WIN );

		adjust_border( MAIN_WIN );
	}

	return( TCL_OK );
}


/* ARGSUSED */
int canvas_handle_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	static int markx = -1;
	static int marky = -1;

	STV_VIEW_PARAM VP;

	double vel_goal;
	double factor;

	char cmd[1024];

	char *op;

	int xmin, xmax;
	int ymin, ymax;

	int x, y;
	int flag;
	int tmp;

	if ( MAIN_WIN == NULL )
		return( TCL_OK );

	if ( argc != 4 )
	{
		Tcl_SetResult( itp, "usage: canvas_handle <op> <x> <y>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	op = argv[1];

	x = atoi( argv[2] );
	y = atoi( argv[3] );

	if ( vflag )
	{
		printf( "(%d, %d) -> (%d, %d)\n", x, y,
			SCREEN_TO_REGION( x, STV_REGION_MIN( MAIN_WIN->region, 0 ),
				MAIN_WIN->xoff, MAIN_WIN->scale ),
			SCREEN_TO_REGION( y, STV_REGION_MIN( MAIN_WIN->region, 1 ),
				MAIN_WIN->yoff, MAIN_WIN->scale ) );
	}

	if ( !strcmp( op, "press" ) )
	{
		if ( PLAY_MODE == PLAYING && CONNECTED )
		{
			PTR_X = x;
			PTR_Y = y;

			/* Select Heli to Steer */

			if ( CURRENT_HELI == NULL )
				request_heli_steer( x, y );

			else
			{
				STEER_RELEASE = stvFalse;

				STEER_ACTIVE = stvTrue;
			}
		}

		else if ( PLAY_MODE == MONITORING )
		{
			/* Create Zoom Rect */

			if ( REGION_ID != -1 )
				DELETE_ITEM( itp, MAIN_WIN->canvas, REGION_ID );
		
			REGION_ID = DRAW_RECT( itp, MAIN_WIN->canvas, x, y, x, y,
				"", "blue", 1 );

			markx = x;
			marky = y;
		}
	}

	else if ( !strcmp( op, "slide" ) )
	{
		if ( PLAY_MODE == PLAYING )
		{
			PTR_X = x;
			PTR_Y = y;
		}

		else if ( PLAY_MODE == MONITORING )
		{
			/* Adjust Zoom Rect */

			if ( REGION_ID != -1 )
			{
				RECT_COORDS( itp, MAIN_WIN->canvas, REGION_ID,
					markx, marky, x, y );
			}
		}
	}

	else if ( !strcmp( op, "release" ) )
	{
		if ( PLAY_MODE == PLAYING )
		{
			PTR_X = x;
			PTR_Y = y;

			STEER_RELEASE = stvTrue;
		}

		else if ( PLAY_MODE == MONITORING )
		{
			/* Set Desired Vis Region */

			DELETE_ITEM( itp, MAIN_WIN->canvas, REGION_ID );

			if ( CONNECTED
				&& markx != x && marky != y
				&& !EQUIV( MAIN_WIN->scale, -1.0 ) )
			{
				/* Save Old Region */

				push_region( MAIN_WIN->region );

				/* Determine Min/Max Coords */

				if ( markx < x )
				{
					xmin = markx;
					xmax = x;
				}

				else
				{
					xmin = x;
					xmax = markx;
				}

				if ( marky < y )
				{
					ymin = marky;
					ymax = y;
				}

				else
				{
					ymin = y;
					ymax = marky;
				}

				/* Set New Region Bounds */

				STV_REGION_MAX( MAIN_WIN->region, 0 ) =
					SCREEN_TO_REGION( xmax,
						STV_REGION_MIN( MAIN_WIN->region, 0 ),
						MAIN_WIN->xoff, MAIN_WIN->scale );

				STV_REGION_MAX( MAIN_WIN->region, 1 ) =
					SCREEN_TO_REGION( ymax,
						STV_REGION_MIN( MAIN_WIN->region, 1 ),
						MAIN_WIN->yoff, MAIN_WIN->scale );

				STV_REGION_MIN( MAIN_WIN->region, 0 ) =
					SCREEN_TO_REGION( xmin,
						STV_REGION_MIN( MAIN_WIN->region, 0 ),
						MAIN_WIN->xoff, MAIN_WIN->scale );

				STV_REGION_MIN( MAIN_WIN->region, 1 ) =
					SCREEN_TO_REGION( ymin,
						STV_REGION_MIN( MAIN_WIN->region, 1 ),
						MAIN_WIN->yoff, MAIN_WIN->scale );

				printf( "New Display Region [%d - %d][%d - %d]\n",
					STV_REGION_MIN( MAIN_WIN->region, 0 ),
					STV_REGION_MAX( MAIN_WIN->region, 0 ),
					STV_REGION_MIN( MAIN_WIN->region, 1 ),
					STV_REGION_MAX( MAIN_WIN->region, 1 ) );

				stv_check_region( CURRENT_VF->field,
					MAIN_WIN->region, VISCELL );

				printf(
					"After stv_check_region():  [%d - %d][%d - %d]\n",
					STV_REGION_MIN( MAIN_WIN->region, 0 ),
					STV_REGION_MAX( MAIN_WIN->region, 0 ),
					STV_REGION_MIN( MAIN_WIN->region, 1 ),
					STV_REGION_MAX( MAIN_WIN->region, 1 ) );

				/* Send New Vis Region */

				set_visregion( stvTrue );

				/* if is_static ever set, check return code here */
				stv_viewer_set_VisRegion( CURRENT_VFG, VIEWER,
					VISREGION, VISCELL );

				/* Set Display Scale & Border */

				set_display_scale( MAIN_WIN );

				adjust_border( MAIN_WIN );

				reset_item_screen_coords( MAIN_WIN );
			}
		}
	}

	else if ( !strcmp( op, "fire" ) )
	{
		if ( PLAY_MODE == PLAYING && CONNECTED )
		{
			PTR_X = x;
			PTR_Y = y;

			/* Fire Bullet */

			if ( CURRENT_HELI != NULL )
			{
				tmp = stvTrue;

				if ( stv_viewer_steer_vparameter( CURRENT_VP,
					(STV_VALUE) &tmp, stvInt, FIRE_BULLET_INDEX )
					!= stvStatusOk )
				{
					printf( "Error steering bullet firing...\n" );
				}

				STEER_ACTIVE = stvTrue;
			}
		}
	}

	else if ( !strcmp( op, "shield" ) )
	{
		if ( PLAY_MODE == PLAYING && CONNECTED )
		{
			PTR_X = x;
			PTR_Y = y;

			/* Toggle Heli Shield */

			if ( CURRENT_HELI != NULL )
			{
				tmp = CURRENT_HELI->shielded ? stvFalse : stvTrue;

				if ( stv_viewer_steer_vparameter( CURRENT_VP,
					(STV_VALUE) &tmp, stvInt, SHIELDED_INDEX )
					!= stvStatusOk )
				{
					printf( "Error steering heli shield...\n" );
				}
			}
		}
	}

	else if ( !strcmp( op, "zoomin" ) )
	{
		if ( PLAY_MODE == PLAYING )
		{
			PTR_X = x;
			PTR_Y = y;

			if ( CURRENT_HELI != NULL && !ZOOM_LIMIT && !AUTO_ZOOM )
			{
				ZOOM_LIMIT = stvTrue;

				ZOOM_FACTOR += 1.0;

				sprintf( cmd, "%s.zoom_factor delete 0 end", FM );
				Tcl_Eval( itp, cmd );

				sprintf( cmd, "%s.zoom_factor insert 0 %lg",
					FM, ZOOM_FACTOR );
				
				Tcl_Eval( itp, cmd );

				set_display_to_heli( CURRENT_HELI, stvTrue );
			}
		}
	}

	else if ( !strcmp( op, "zoomout" ) )
	{
		if ( PLAY_MODE == PLAYING )
		{
			PTR_X = x;
			PTR_Y = y;

			if ( CURRENT_HELI != NULL && !AUTO_ZOOM )
			{
				ZOOM_FACTOR -= 1.0;

				if ( ZOOM_FACTOR < 1.0 )
					ZOOM_FACTOR = 1.0;

				sprintf( cmd, "%s.zoom_factor delete 0 end", FM );
				Tcl_Eval( itp, cmd );

				sprintf( cmd, "%s.zoom_factor insert 0 %lg",
					FM, ZOOM_FACTOR );
				
				Tcl_Eval( itp, cmd );

				set_display_to_heli( CURRENT_HELI, stvTrue );
			}
		}
	}

	else if ( !strcmp( op, "pop" ) )
	{
		if ( PLAY_MODE == MONITORING
			&& CONNECTED && pop_region( MAIN_WIN->region ) )
		{
			set_visregion( stvTrue );

			/* if is_static ever set, check return code here */
			stv_viewer_set_VisRegion( CURRENT_VFG, VIEWER,
				VISREGION, VISCELL );

			set_display_scale( MAIN_WIN );

			adjust_border( MAIN_WIN );

			reset_item_screen_coords( MAIN_WIN );
		}
	}

	else if ( !strcmp( op, "POP" ) )
	{
		if ( PLAY_MODE == MONITORING && CONNECTED )
		{
			flag = 0;

			while ( pop_region( MAIN_WIN->region ) )
				flag++;

			if ( flag )
			{
				set_visregion( stvTrue );

				/* if is_static ever set, check return code here */
				stv_viewer_set_VisRegion( CURRENT_VFG, VIEWER,
					VISREGION, VISCELL );

				set_display_scale( MAIN_WIN );

				adjust_border( MAIN_WIN );

				reset_item_screen_coords( MAIN_WIN );
			}
		}
	}

	else if ( !strcmp( op, "enter" ) )
	{
		if ( PLAY_MODE == PLAYING )
		{
			PTR_VALID = stvTrue;

			PTR_X = x;
			PTR_Y = y;
		}
	}

	else if ( !strcmp( op, "move" ) )
	{
		if ( PLAY_MODE == PLAYING )
		{
			PTR_X = x;
			PTR_Y = y;
		}
	}

	else if ( !strcmp( op, "leave" ) )
	{
		if ( PLAY_MODE == PLAYING )
		{
			PTR_VALID = stvFalse;

			PTR_X = -1;
			PTR_Y = -1;
		}
	}

	else if ( strlen( op ) == 1 )
	{
		switch ( op[0] )
		{
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			{
				if ( PLAY_MODE == PLAYING )
				{
					VP = stv_get_view_param_name( "heli velocity",
						VIEWER->vparams, VIEWER->nvparams );

					if ( VP != NULL )
					{
						factor =
							STV_VALUE_OF( VP->param->vars[0], double )
								/ (double) 9.0;
					}

					else
						factor = 1.0;

					vel_goal = factor * atof( op );

					if ( stv_viewer_steer_vparameter( CURRENT_VP,
						(STV_VALUE) &vel_goal, stvDouble,
						VELOCITY_GOAL_INDEX ) != stvStatusOk )
					{
						printf( "Error steering heli velocity...\n" );
					}
				}

				break;
			}

			case ' ':
			{
				if ( PLAY_MODE == PLAYING && CONNECTED )
				{
					PTR_X = x;
					PTR_Y = y;

					/* Fire Bullet */

					if ( CURRENT_HELI != NULL )
					{
						tmp = stvTrue;

						if ( stv_viewer_steer_vparameter( CURRENT_VP,
							(STV_VALUE) &tmp, stvInt,
							FIRE_BULLET_INDEX ) != stvStatusOk )
						{
							printf(
								"Error steering bullet firing...\n" );
						}

						STEER_ACTIVE = stvTrue;
					}
				}

				break;
			}

			case 'm':
			{
				SET_TCL_GLOBAL( itp, "play_mode", "Monitor" );

				Tcl_Eval( itp, "layout_main_panel" );

				Tcl_Eval( itp, "set_play_mode Monitor" );

				break;
			}

			case 'p':
			{
				PTR_VALID = stvTrue;

				PTR_X = x;
				PTR_Y = y;

				SET_TCL_GLOBAL( itp, "play_mode", "Play" );

				Tcl_Eval( itp, "layout_main_panel" );

				Tcl_Eval( itp, "set_play_mode Play" );

				request_heli_steer( x, y );

				break;
			}

			case 's':
			{
				qflag = qflag ? stvFalse : stvTrue;

				printf( "Sound Toggled %s\n",
					qflag ? "OFF" : "ON" );

				break;
			}

			case 'q':
			case 'Q':
			{
				Tcl_Eval( itp, "quit" );

				break;
			}

			default:
			{
				printf( "Error in canvas_handle{}: Unknown Key '%c'\n",
					op[0] );
				
				break;
			}
		}
	}

	else
		printf( "Error in canvas_handle{}: Unknown Op = \"%s\"\n", op );

	return( TCL_OK );
}


void
request_heli_steer( x, y )
int x, y;
{
	char cmd[1024];

	int tmp;
	int cc;

	/* Find Closest Heli */

	CURRENT_HELI = closest_heli( MAIN_WIN, x, y,
		(HELI) NULL, (double *) NULL );

	/* Try to Steer it */

	if ( CURRENT_HELI != NULL )
	{
		if ( !STEERING )
		{
			printf( "Initializing Steering\n" );

			cc = stv_viewer_steering_init( VIEWER, 1, stvFalse,
				&RESTART );
			
			if ( cc == stvStatusOk )
				STEERING = stvTrue;
			
			else
			{
				printf( "*** Steering Initialization Failed ***\n" );

				CURRENT_HELI = (HELI) NULL;
			}
		}

		if ( STEERING )
		{
			printf( "Requesting Helicopter hid=%d\n",
				CURRENT_HELI->hid );

			CURRENT_VP = stv_get_view_param_name( "helicopter",
				VIEWER->vparams, VIEWER->nvparams );

			if ( CURRENT_VP == NULL )
			{
				printf( "\nError Locating Heli Param\n\n" );

				CURRENT_HELI = (HELI) NULL;
			}
			
			else
			{
				/* Set Steering Index */

				if ( stv_viewer_steer_vparameter( CURRENT_VP,
					(STV_VALUE) &(CURRENT_HELI->hid), stvInt,
					HELI_INDEX ) != stvStatusOk )
				{
					CURRENT_HELI = (HELI) NULL;

					return;
				}

				/* Request Heli Steer */

				stv_viewer_steering_request( CURRENT_VP, VIEWER,
					stvMsgNoBlock );

				if ( CURRENT_VP->token != stvSteerToken )
				{
					printf( "Steering Token Not Available\n" );

					stv_viewer_steering_release( CURRENT_VP, VIEWER );

					CURRENT_HELI = (HELI) NULL;
				}

				/* Fill In & Init Current Param Values */
				/* (also set vis region for heli radar) */

				else
				{
					printf( "Steering Token Obtained\n" );

					/* Heading */

					stv_copy_converted_data(
						&(CURRENT_HELI->heading), stvDouble,
						CURRENT_VP->param->vars[ HEADING_GOAL_INDEX ],
						CURRENT_VP->param->types[ HEADING_GOAL_INDEX ]
							& stvTypeMask );

					/* Velocity */

					stv_copy_converted_data(
						&(CURRENT_HELI->velocity), stvDouble,
						CURRENT_VP->param->vars[ VELOCITY_GOAL_INDEX ],
						CURRENT_VP->param->types[ VELOCITY_GOAL_INDEX ]
							& stvTypeMask );

					/* Radar Range */

					stv_copy_converted_data(
						&(CURRENT_HELI->radar_range), stvDouble,
						CURRENT_VP->param->vars[ RADAR_RANGE_INDEX ],
						CURRENT_VP->param->types[ RADAR_RANGE_INDEX ]
							& stvTypeMask );

					/* Fire Bullet */

					tmp = stvFalse;

					stv_copy_converted_data( &tmp, stvInt,
						CURRENT_VP->param->vars[ FIRE_BULLET_INDEX ],
						CURRENT_VP->param->types[ FIRE_BULLET_INDEX ]
							& stvTypeMask );

					/* Shields */

					tmp = CURRENT_HELI->shielded;

					stv_copy_converted_data( &tmp, stvInt,
						CURRENT_VP->param->vars[ SHIELDED_INDEX ],
						CURRENT_VP->param->types[ SHIELDED_INDEX ]
							& stvTypeMask );

					/* Release */

					tmp = stvFalse;

					stv_copy_converted_data( &tmp, stvInt,
						CURRENT_VP->param->vars[ RELEASE_INDEX ],
						CURRENT_VP->param->types[ RELEASE_INDEX ]
							& stvTypeMask );

					/* Send Dummy Steer to Region */

					CURRENT_VP->changed = stvTrue;

					stv_viewer_send_NewParams( VIEWER );

					/* Set Radar Range Interface */

					RADAR_RANGE = CURRENT_HELI->radar_range;

					sprintf( cmd, "%s.radar_range delete 0 end", FM );
					Tcl_Eval( interp, cmd );

					sprintf( cmd, "%s.radar_range insert 0 %lg",
						FM, CURRENT_HELI->radar_range );
				
					Tcl_Eval( interp, cmd );

					/* Set Score Interface */

					sprintf( cmd,
						"%s.score configure -text \"Score:%10d\"",
						FM_I, CURRENT_HELI->score );

					Tcl_Eval( interp, cmd );

					BASE_SCORE = CURRENT_HELI->score;

					/* Set Vis Region to Heli's Radar */

					push_region( MAIN_WIN->region );

					set_display_to_heli( CURRENT_HELI, stvTrue );

					reset_item_screen_coords( MAIN_WIN );

					activate_long_range_radar();
				}
			}
		}
	}
}


/* ARGSUSED */
int region_slide_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char tmp[1024];

	char *op;

	int *glb, *gub;

	int xdelta, ydelta;
	int xdist, ydist;
	int xfix, yfix;

	if ( argc != 2 )
	{
		Tcl_SetResult( itp,
			"usage: region_slide ( up | down | left | right )",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	if ( PLAY_MODE != MONITORING || !CONNECTED )
		return( TCL_OK );

	op = argv[1];

	if ( !strcmp( op, "up" ) )
	{
		xdelta = 0;
		ydelta = -1;
	}

	else if ( !strcmp( op, "down" ) )
	{
		xdelta = 0;
		ydelta = 1;
	}

	else if ( !strcmp( op, "left" ) )
	{
		xdelta = -1;
		ydelta = 0;
	}

	else if ( !strcmp( op, "right" ) )
	{
		xdelta = 1;
		ydelta = 0;
	}

	else if ( !strcmp( op, "ne" ) )
	{
		xdelta = 1;
		ydelta = -1;
	}

	else if ( !strcmp( op, "nw" ) )
	{
		xdelta = -1;
		ydelta = -1;
	}

	else if ( !strcmp( op, "sw" ) )
	{
		xdelta = -1;
		ydelta = 1;
	}

	else if ( !strcmp( op, "se" ) )
	{
		xdelta = 1;
		ydelta = 1;
	}

	else
	{
		sprintf( tmp, "region_slide{}: Unknown Operation \"%s\"", op );

		Tcl_SetResult( itp, copy_str( tmp ), TCL_DYNAMIC );

		return( TCL_ERROR );
	}

	xdist = STV_REGION_MAX( MAIN_WIN->region, 0 )
		- STV_REGION_MIN( MAIN_WIN->region, 0 ) + 1;

	ydist = STV_REGION_MAX( MAIN_WIN->region, 1 )
		- STV_REGION_MIN( MAIN_WIN->region, 1 ) + 1;

	xdelta *= xdist / 4;
	ydelta *= ydist / 4;

	STV_REGION_MAX( MAIN_WIN->region, 0 ) += xdelta;
	STV_REGION_MIN( MAIN_WIN->region, 0 ) += xdelta;

	STV_REGION_MAX( MAIN_WIN->region, 1 ) += ydelta;
	STV_REGION_MIN( MAIN_WIN->region, 1 ) += ydelta;

	/* Check for Edge of Comp Region */

	xfix = yfix = 0;

	glb = CURRENT_VF->field->particle->glb;
	gub = CURRENT_VF->field->particle->gub;

	if ( STV_REGION_MIN( MAIN_WIN->region, 0 ) < glb[0] )
		xfix = glb[0] - STV_REGION_MIN( MAIN_WIN->region, 0 );

	else if ( STV_REGION_MAX( MAIN_WIN->region, 0 ) > gub[0] )
		xfix = gub[0] - STV_REGION_MAX( MAIN_WIN->region, 0 );

	if ( STV_REGION_MIN( MAIN_WIN->region, 1 ) < glb[1] )
		yfix = glb[1] - STV_REGION_MIN( MAIN_WIN->region, 1 );

	else if ( STV_REGION_MAX( MAIN_WIN->region, 1 ) > gub[1] )
		yfix = gub[1] - STV_REGION_MAX( MAIN_WIN->region, 1 );

	/* Apply Edge Fix */

	STV_REGION_MAX( MAIN_WIN->region, 0 ) += xfix;
	STV_REGION_MIN( MAIN_WIN->region, 0 ) += xfix;

	STV_REGION_MAX( MAIN_WIN->region, 1 ) += yfix;
	STV_REGION_MIN( MAIN_WIN->region, 1 ) += yfix;

	/* Check for NOP */

	if ( xdelta + xfix != 0 || ydelta + yfix != 0 )
	{
		/* Send New Vis Region */

		set_visregion( stvTrue );

		/* if is_static ever set, check return code here */
		stv_viewer_set_VisRegion( CURRENT_VFG, VIEWER,
			VISREGION, VISCELL );

		/* Set Display Scale & Border */

		set_display_scale( MAIN_WIN );

		adjust_border( MAIN_WIN );

		reset_item_screen_coords( MAIN_WIN );
	}

	return( TCL_OK );
}


/* ARGSUSED */
int register_menuvar_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	/*
	char *menuvar;
	char *index;
	char *menu;
	*/

	if ( argc != 4 )
	{
		Tcl_SetResult( itp,
			"usage: register_menuvar <menu> <index> <menuvar>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	/*
	menu	= argv[1];
	index	= argv[2];
	menuvar	= argv[3];
	*/

	/* if ( !strcmp( menu, ".matrix.attr_list_menu" ) ) */

	return( TCL_OK );
}


/* ARGSUSED */
int load_bitmap_file_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	Pixmap bitmap;

	char tmp[1024];

	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: load_bitmap_file <file>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	sprintf( tmp, "@%s", argv[1] );

	bitmap = Tk_GetBitmap( itp, Top, Tk_GetUid( tmp ) );

	if ( bitmap <= 0 )
	{
		sprintf( tmp, "Error Opening Bitmap %s", argv[1] );

		Tcl_SetResult( itp, copy_str( tmp ), TCL_DYNAMIC );

		return( TCL_ERROR );
	}

	return( TCL_OK );
}


/* ARGSUSED */
int fix_help_line_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char tmp[1024];

	char *result;
	char *ptr;
	char *str;

	int len;

	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: fix_help_line <line>", TCL_STATIC );

		return( TCL_ERROR );
	}

	str = argv[1];

	ptr = tmp;

	while ( *str != '\0' )
	{
		if ( *str == '' )
			*ptr++ = '{';

		else if ( *str == '' )
			*ptr++ = '}';

		else
			*ptr++ = *str;

		str++;
	}

	*ptr = '\0';

	result = copy_str( tmp );

	Tcl_SetResult( itp, result, TCL_DYNAMIC );

	return( TCL_OK );
}


/* ARGSUSED */
int double_click_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	static struct timeval last = { -1, -1 };

	struct timeval tmp;

	int diff;

	gettimeofday( &tmp, (struct timezone *) NULL );

	diff = ( ( tmp.tv_sec - last.tv_sec ) * 1000000 )
		+ tmp.tv_usec - last.tv_usec;

	if ( diff < 500000 )
		Tcl_SetResult( itp, "1", TCL_STATIC );

	else
		Tcl_SetResult( itp, "0", TCL_STATIC );

	last.tv_sec = tmp.tv_sec;
	last.tv_usec = tmp.tv_usec;

	return( TCL_OK );
}


/* ARGSUSED */
int strip_label_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char tmp[1024];

	char *result;
	char *ptr;
	char *str;

	int len;

	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: strip_label <str>", TCL_STATIC );

		return( TCL_ERROR );
	}

	str = argv[1];

	ptr = tmp;

	while ( *str != '\0' )
	{
		if ( *str == '-' || *str == ' ' || *str == '.' )
			*ptr++ = '_';

		else
			*ptr++ = *str;

		str++;
	}

	*ptr = '\0';

	result = copy_str( tmp );

	Tcl_SetResult( itp, result, TCL_DYNAMIC );

	return( TCL_OK );
}


/* ARGSUSED */
int title_info_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char tmp[255];

	sprintf( tmp, "%s", FLYME_VERSION );

	Tcl_SetResult( itp, tmp, TCL_STATIC );

	return( TCL_OK );
}


/* ARGSUSED */
int pad_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char *result;
	char *ptr;
	char *str;

	int len;
	int nd;
	int nz;
	int i;

	if ( argc < 3 )
	{
		Tcl_SetResult( itp,
			"usage: pad <number_str> <desired_length>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	str = argv[1];

	len = atoi( argv[2] );

	nd = strlen( str );

	nz = len - nd;

	result = (char *) malloc( (STV_SIZE_T) (len + 1) * sizeof(char) );
	memcheck( result, "Numerical Pad String" );

	ptr = result;

	for ( i=0 ; i < nz ; i++ )
		*ptr++ = '0';

	sprintf( (char *) (result + nz), "%s", str );

	Tcl_SetResult( itp, result, TCL_DYNAMIC );

	return( TCL_OK );
}


/* ARGSUSED */
void
quit_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	release_heli();

	stv_viewer_send_FieldHaltAll( CURRENT_VFG, VIEWER );

	exit( 0 );
}


