
/* $Id: slicer.h,v 1.17 1999/11/03 01:33:18 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <sys/types.h>

#ifndef WIN32
#include <pwd.h>
#endif

#include <math.h>


#ifndef PI
#define PI (M_PI)
#endif


/* STV Header Files */

#include "stv.h"


/* TCL / TK Header Files */

#include <tk.h>


/* Slicer Constants & Macros */

#include "const.h"


/* Data Structures */

struct rect_struct
{
	int x, y, z;
	int x1, y1;
	int x2, y2;
	int id;
	float value;
	char color[16];
};

typedef struct rect_struct *RECT;

struct circle_struct
{
	int x, y, z;
	int xc, yc;
	int r;
	int id;
	float radius;
	float value;
	char color[16];
};

typedef struct circle_struct *CIRCLE;

struct win_struct
{
	char *canvas;
	char *htvar, *wtvar;
	double scale;
	int csx, csy;
	STV_REGION region;
	int xoff, yoff;
	int border_id;
};

typedef struct win_struct *WIN;


struct region_stack_struct
{
	STV_REGION region;
	/* REGION_STACK */ struct region_stack_struct *next;
};

typedef struct region_stack_struct *REGION_STACK;


typedef void (*vfp)();

typedef int (*ifp)();


/* Routines */

STV_VIEW_FIELD select_field();

CIRCLE	find_circle();

RECT	find_rect();

WIN		create_win();

FILE	*fopen();

double	distance();
double	angle_of();
double	atof();

char	*upper_str();
char	*trunc_str();
char	*pad_num();

void	color_value();
void	do_stv_cleanup();
void	free_region_stack();
void	free_win();
void	handle_frame();
void	handle_min_max();
void	load_current_field();
void	memcheck();
void 	push_region();
void	redraw_view();
void	safe_eval();
void	set_current_field();
void	save_current_field();
void	update_color_range();
void	update_vis_region();
void	viewer_handle();

int		viewer_lock();
int		viewer_unlock();

