
/* $Id: apptest.c,v 3.16 1999/04/15 16:15:31 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <malloc.h>
#include <math.h>


/* STV Header Files */

#include "stv.h"


/* PVM Header Files */

#include "pvm3.h"


/* Message Types */

#define msgSecondaryNotify    -1
#define msgReconfigure        -2


/* Macros */

#define GET_DATA_TYPE( _num ) \
	( !(_num) ? stvInt : \
		( (_num) == 1 ? stvFloat : \
			stvDouble ) )

#define GET_AXIS_TYPE( _num ) \
	( !(_num) ? stvBlock : \
		( (_num) == 1 ? stvCyclic : \
			( (_num) == 2 ? stvExplicit : \
				stvCollapse ) ) )

#define DIM_SIZE( _decomp, _index ) \
	( (_decomp)->gub[ _index ] - (_decomp)->glb[ _index ] + 1 )


/* Max Array Sizes */

#define MAX_DECOMP 100
#define MAX_FIELD  100
#define MAX_PARAM  100
#define MAX_PART   100


/* Structures */

struct decomp_info
{
	int id;
	int data_dim;
	int axis_type[STV_MAX_DIM];
	int axis_info1[STV_MAX_DIM];
	int axis_info2[STV_MAX_DIM];
	int *axis_bounds[STV_MAX_DIM];
	int *axis_bounds2[STV_MAX_DIM];
	int nbounds[STV_MAX_DIM];
	int glb[STV_MAX_DIM];
	int gub[STV_MAX_DIM];
	int ntasks;
	int prank;
	int pshape[STV_MAX_DIM];
};

struct part_struct
{
	int coords[STV_MAX_DIM];
	double foo_double;
	float foo_float[2];
	int foo_int[4];
	struct part_struct *next;
};

typedef struct part_struct *PART;

PART create_part();


/* Globals */

struct decomp_info decomps[MAX_DECOMP];

STV_VALUE varptrs[MAX_FIELD];

PART PART_LIST;
int NPARTS;

void	get_part __STV_Prototype__(( int, STV_REGION,
			STV_PARTICLE_ID *, int * ));

void	get_foo_pfield __STV_Prototype__(( STV_PARTICLE_ID,
			STV_VALUE, STV_VALUE * ));


void	make_part __STV_Prototype__(( int *, STV_PARTICLE_ID * ));

void	set_foo_pfield __STV_Prototype__(( STV_PARTICLE_ID,
			STV_VALUE, STV_VALUE ));


#define PF_DOUBLE	0
#define PF_FLOAT	1
#define PF_INT		2


double dparameters[MAX_PARAM];
float fparameters[MAX_PARAM];
int iparameters[MAX_PARAM];

char *APPNAME;

int NDECOMPS;
int NPARAMS;
int NFIELDS;

int NOTIFY;
int MSGTAG;
int NTASKS;
int NSPAWN;

int SIZE;

int SPMD;

int myinst;
int mytid;

int vflag;
int dflag;


/* Routines */

char	**add_argv();

void	check_decomp_info();
void	clear_stv_stuff();
void	do_stv_stuff();
void	fill_var_array();
void	free_part();
void	program_init();
void	read_args();
void	set_paddr();
void	set_pshape();
void	usage();


/* Main */

main( argc, argv )
int argc;
char **argv;
{
	char **restart_argv;

	char *appgroup;

	int param_ids[MAX_PARAM];
	int field_ids[MAX_FIELD];
	int part_ids[MAX_PART];

	int nfields;
	int budinst;
	int buddy;
	int flags;
	int done;
	int tid;
	int tmp;

	read_args( argc, argv );

	program_init();

	NOTIFY = MSGTAG + 1;

	mytid = pvm_mytid();

	if ( mytid < 0 )
	{
		pvm_perror( "Obtaining My TID" );

		exit( 0 );
	}

	printf( "My TID = [ 0x%x / %d ]\n", mytid, mytid );
	fflush( stdout );

	if ( APPNAME != NULL )
		appgroup = APPNAME;

	else
		appgroup = "renumber_me";

	myinst = pvm_joingroup( appgroup );

	if ( !SPMD )
	{
		if ( !myinst && NTASKS > 1 )
		{
			printf( "Spawning %d Other Tasks... ", NTASKS - 1 );
			fflush( stdout );

			if ( dflag )
				flags = PvmTaskDefault | PvmTaskDebug;

			else
				flags = PvmTaskDefault;

			pvm_spawn( "apptest", argv + 1, flags,
				(char *) NULL, NTASKS - 1, (int *) NULL );

			printf( "done.\n" );
			fflush( stdout );
		}
	}

	/* Go for it */

	while ( 1 )
	{
		/* Set Up Ring Notifies */

		budinst = ( myinst + 1 ) % NTASKS;

		while ( (buddy = pvm_gettid( appgroup, budinst )) < 0 );

		pvm_notify( PvmTaskExit, NOTIFY, 1, &buddy );

		/* (Re)Initialize STV Stuff */

		clear_stv_stuff( field_ids, &nfields, param_ids, part_ids );

		do_stv_stuff( field_ids, &nfields, param_ids, part_ids );

		/* Iterate */

		done = 0;

		do
		{
			/* Check for Notify */

			if ( pvm_nrecv( -1, NOTIFY ) )
			{
				printf( "\nGOT NOTIFY\n" );

				pvm_upkint( &tid, 1, 1 );

				/* Buddy Task Died, Handle Regeneration */

				if ( tid > 0 )
				{
					printf( "\tTask TID=[ 0x%x / %d ] Has EXITED.\n",
						tid, tid );
					fflush( stdout );

					if ( tid == buddy )
					{
						/* Broadcast Secondary Notifies */

						if ( NSPAWN == 1 )
						{
							printf( "\tSending Secondary Notify.\n" );
							fflush( stdout );

							pvm_initsend( PvmDataDefault );

							tmp = msgSecondaryNotify;
							pvm_pkint( &tmp, 1, 1 );

							pvm_bcast( appgroup, NOTIFY );
						}

						/* Broadcast New Configuration */

						else
						{
							printf( "\tSending New Configuration.\n" );
							fflush( stdout );

							NTASKS += NSPAWN - 1;

							pvm_initsend( PvmDataDefault );

							tmp = msgReconfigure;
							pvm_pkint( &tmp, 1, 1 );

							pvm_pkint( &NTASKS, 1, 1 );

							pvm_bcast( appgroup, NOTIFY );
						}

						/* Restart Dead Task */

						printf( "\tSpawning %d Other Tasks... ",
							NSPAWN );
						fflush( stdout );

						restart_argv = add_argv( argv + 1, argc - 1,
							"-SPMD" );

						pvm_spawn( "apptest", restart_argv,
							PvmTaskDefault, (char *) NULL, NSPAWN,
							(int *) NULL );

						free( restart_argv );

						printf( "done.\n" );
						fflush( stdout );

						done++;
					}

					else
					{
						printf( "\tSo Who Cares?  Drop it.\n" );
						fflush( stdout );
					}
				}

				/* Reconfigure Notification */

				else if ( tid == msgReconfigure )
				{
					printf( "\tReconfigure Notification:\n" );
					fflush( stdout );

					pvm_upkint( &NTASKS, 1, 1 );

					printf( "\tNew NTASKS = %d\n", NTASKS );
					fflush( stdout );

					done++;
				}

				/* Secondary Notification Only */

				else if ( tid == msgSecondaryNotify )
				{
					printf( "\tSecondary Notification, Restart.\n" );
					fflush( stdout );

					done++;
				}
			}

			else
			{
				/* STV Stuff */

				if ( vflag )
				{
					printf( "Calling stv_sendToFE()...\n" );
					fflush( stdout );
				}

				stv_sendToFE();

				if ( vflag )
				{
					printf( "Returned from stv_sendToFE().\n" );
					fflush( stdout );
				}

				stvSleep( 1 );
			}
		}
		while ( !done );
	}
}


void
do_stv_stuff( field_ids, nfields, param_ids, part_ids )
int *field_ids;
int *nfields;
int *param_ids;
int *part_ids;
{
	struct decomp_info *D;

	STV_VALUE ptr;

	PART P;

	char tmp[1024];

	unsigned sz;

	int array_decl[STV_MAX_DIM];
	int offsets[STV_MAX_DIM];
	int paddr[STV_MAX_DIM];
	int glb[STV_MAX_DIM];
	int gub[STV_MAX_DIM];

	int type;
	int i, j;
	int num;
	int p;

	/* Initialize STV Info */

	stv_setopt( stvOptNull, stvOptVerbose, vflag ? STV_DEBUG_ALL : 0 );

	stv_init( APPNAME, MSGTAG, NTASKS, myinst );

	/* Define Parameters */

	for ( i=0 ; i < NPARAMS ; i++ )
	{
		sprintf( tmp, "param #%d", i + 1 );

		type = GET_DATA_TYPE( i % 3 );

		if ( type == stvInt )
		{
			iparameters[i] = i;

			ptr = (STV_VALUE) &(iparameters[i]);
		}

		else if ( type == stvFloat )
		{
			fparameters[i] = (float) i;

			ptr = (STV_VALUE) &(fparameters[i]);
		}

		else if ( type == stvDouble )
		{
			dparameters[i] = (double) i;

			ptr = (STV_VALUE) &(dparameters[i]);
		}

		printf( "Defining Parameter #%d:  name=\"%s\" type=%s\n",
			i, tmp, stv_data_type_str( type ) );
		fflush( stdout );

		param_ids[i] = stv_paramDefine( tmp, ptr, type,
			stvVisOnly | stvCpOnly );
	}

	/* Define Decompositions */

	for ( i=0 ; i < NDECOMPS ; i++ )
	{
		printf( "Defining Decomposition #%d:\n", i );
		fflush( stdout );

		D = &(decomps[i]);

		/* Decomposition Dimension */

		D->data_dim = ( i % 3 ) + 1;

		printf( "\tdata dim = %d\n", D->data_dim );
		fflush( stdout );

		/* Number of Processors */

		D->ntasks = NTASKS - i;

		while ( D->ntasks < 1 )
			D->ntasks += NTASKS;

		printf( "\tntasks = %d\n", D->ntasks );
		fflush( stdout );

		/* Global Bounds, Axis Types & Info */

		D->prank = 0;

		for ( j=0 ; j < D->data_dim ; j++ )
		{
			/* Global Array Lower & Upper Bounds */

			D->glb[j] = i * ( j + 1 );

			D->gub[j] = D->glb[j] + ( ( j + 1 ) * SIZE ) - 1;

			/* Decomposition Type */

			D->axis_type[j] = GET_AXIS_TYPE( ( i + j ) % 3 );

			/* Decomposition Info */

			if ( D->axis_type[j] == stvCollapse )
			{
				D->axis_info1[j] = -1;
				D->axis_info2[j] = -1;
			}

			else if ( D->axis_type[j] == stvExplicit )
			{
				D->axis_bounds[j] = (int *) malloc( sizeof( int ) );
				D->axis_bounds2[j] = (int *) malloc( sizeof( int ) );

				if ( stv_memcheck( D->axis_bounds[j],
							"Axis Bounds Sub-Array" ) != stvStatusOk
						|| stv_memcheck( D->axis_bounds2[j],
							"Axis Bounds Sub-Array" ) != stvStatusOk )
					exit( -1 );

				D->axis_bounds[j][0] = -1;
				D->axis_bounds2[j][0] = -1;

				D->nbounds[j] = 1;

				D->axis_info1[j] = -1;
				D->axis_info2[j] = -1;

				(D->prank)++;
			}

			else
			{
				D->axis_info1[j] = ( i * j ) ? ( i * j ) : -1;

				D->axis_info2[j] = -1;

				(D->prank)++;
			}

			/* Check Validity of Parameters */

			stv_check_decomp_axis_info(
				D->axis_type[j], &(D->axis_info1[j]),
				D->axis_bounds[j], D->axis_bounds2[j], D->nbounds[j],
				D->glb[j], D->gub[j], D->ntasks );

			printf(
				"\t\t%s [%d]: type=%s %s=%d,%d glb=%d gub=%d\n",
				"Initial", j, stv_decomp_type_str( D->axis_type[j] ),
				"info", D->axis_info1[j], D->axis_info2[j],
				D->glb[j], D->gub[j] );
			fflush( stdout );
		}

		/* Processor Rank & Shape */

		printf( "\tprank = %d\n", D->prank );
		fflush( stdout );

		set_pshape( D->prank, D->pshape, D->ntasks );

		for ( j=0 ; j < D->prank ; j++ )
			printf( "\t\t[%d] = %d\n", j, D->pshape[j] );
		
		fflush( stdout );

		/* Verify Proper Axis Info vs. Processor Shape */

		check_decomp_info( D );

		for ( j=0 ; j < D->data_dim ; j++ )
		{
			if ( D->axis_type[j] == stvExplicit )
			{
				printf(
		"\t\t%s [%d]: type=%s %s=%d,%d %s=%d,%d(%d) glb=%d gub=%d\n",
					"Final", j, stv_decomp_type_str( D->axis_type[j] ),
					"info", D->axis_info1[j], D->axis_info2[j],
					"bounds", D->axis_bounds[j][0],
					D->axis_bounds2[j][0], D->nbounds[j],
					D->glb[j], D->gub[j] );
			}

			else
			{
				printf(
					"\t\t%s [%d]: type=%s %s=%d,%d glb=%d gub=%d\n",
					"Final", j, stv_decomp_type_str( D->axis_type[j] ),
					"info", D->axis_info1[j], D->axis_info2[j],
					D->glb[j], D->gub[j] );
			}

			fflush( stdout );
		}

		/* Define STV Decomposition */

		D->id = stv_decompDefine( D->data_dim,
			D->axis_type, D->axis_info1, D->axis_info2, D->glb, D->gub,
			D->prank, D->pshape );
	}

	/* Clear "offsets" Array */

	for ( i=0 ; i < STV_MAX_DIM ; i++ )
		offsets[i] = 0;

	/* Define Fields */

	*nfields = 0;

	for ( i=0 ; i < NFIELDS ; i++ )
	{
		D = &(decomps[ i % NDECOMPS ]);

		p = myinst - ( i % NTASKS );

		if ( p < 0 )
			p += NTASKS;

		if ( p < D->ntasks )
		{
			/* Field Name */

			sprintf( tmp, "field #%d", i );

			printf(
				"Defining Field #%d (\"%s\") with p=%d, decomp #%d:\n",
				i, tmp, p, i % NDECOMPS );
			fflush( stdout );

			/* Processor Address */

			set_paddr( D->prank, D->pshape, p, paddr );

			printf( "\tpaddr:\n" );

			for ( j=0 ; j < D->prank ; j++ )
				printf( "\t\t[%d] = %d\n", j, paddr[j] );
			
			fflush( stdout );

			/* Data Type */

			type = GET_DATA_TYPE( i % 3 );

			printf( "\ttype = %s\n", stv_data_type_str( type ) );
			fflush( stdout );

			if ( type == stvInt )
				sz = sizeof( int );
			
			else if ( type == stvFloat )
				sz = sizeof( float );

			else if ( type == stvDouble )
				sz = sizeof( double );

			/* Allocate Variable Storage */

			num = get_var_alloc_size( D, paddr, array_decl );

			if ( num < 0 )
				return;

			printf( "\tallocated elements = %d:\n", num );

			for ( j=0 ; j < D->data_dim ; j++ )
				printf( "\t\tarray_decl[%d] = %d\n", j, array_decl[j] );
			
			fflush( stdout );

			if ( num > 0 )
			{
				/* Allocate Variable Storage */

				varptrs[ *nfields ] = (STV_VALUE) malloc(
					(STV_SIZE_T) num * sz );

				if ( stv_memcheck( varptrs[ *nfields ],
					"Variable Allocation" ) != stvStatusOk )
				{
					exit( -1 );
				}

				/* Initialize Local Array Values */

				ptr = varptrs[ *nfields ];

				fill_var_array( ptr, D, array_decl, type, paddr );

				if ( vflag )
				{
					printf( "Local Field Array:\n" );
					fflush( stdout );

					stv_dump_user_field_array( ptr, array_decl,
						D->data_dim, offsets, type, stvColumnMajor );
				}
			}

			else
				ptr = varptrs[ *nfields ] = (STV_VALUE) NULL;

			/* Define STV Field */

			field_ids[ *nfields ] = stv_fieldDefine( ptr, tmp, D->id,
				stvNoFieldOffsets, array_decl, type, paddr,
				stvVisOnly | stvCpOnly );

			(*nfields)++;
		}
	}

	/* Define Particles */

	glb[0] = 1;
	glb[1] = 10;
	glb[2] = 100;

	gub[0] = ( 1000 * NTASKS ) + glb[0] - 1;
	gub[1] = 1000 + glb[1] - 1;
	gub[2] = 1000 + glb[2] - 1;

	for ( i=0 ; i < SIZE ; i++ )
	{
		if ( PART_LIST == NULL )
			P = PART_LIST = create_part();

		else
			P = P->next = create_part();

		P->coords[0] = ( i * ( 1000 / SIZE ) )
			+ ( 1000 * myinst ) + glb[0];
		
		P->coords[1] = ( i * ( 1000 / SIZE ) ) + glb[1];
		P->coords[2] = ( 1000 - ( i * ( 1000 / SIZE ) ) ) + glb[2] - 1;

		P->foo_double = ( 100.0 * (double) i )
			+ ( 1000.0 * (double) myinst );

		for ( j=0 ; j < 2 ; j++ )
		{
			P->foo_float[j] = ( 10.0 * (float) i )
				+ ( 1000.0 * (float) myinst ) + j;
		}

		for ( j=0 ; j < 4 ; j++ )
			P->foo_int[j] = i + j + ( 1000 * myinst );
	}

	part_ids[0] = stv_particleDefine( "foo", 3, glb, gub, NTASKS,
		get_part, make_part );

	stv_pfieldDefine( "foo_double", part_ids[0], stvDouble, 1,
		get_foo_pfield, (STV_VALUE) PF_DOUBLE,
		set_foo_pfield, (STV_VALUE) PF_DOUBLE,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "foo_float", part_ids[0], stvFloat, 2,
		get_foo_pfield, (STV_VALUE) PF_FLOAT,
		set_foo_pfield, (STV_VALUE) PF_FLOAT,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "foo_int", part_ids[0], stvInt, 4,
		get_foo_pfield, (STV_VALUE) PF_INT,
		set_foo_pfield, (STV_VALUE) PF_INT,
		stvVisOnly | stvCpOnly );
}


void
set_pshape( prank, pshape, ntasks )
int prank;
int *pshape;
int ntasks;
{
	int factor;
	int np;
	int i;

	/* Initial Identity Shape */

	for ( i=0 ; i < prank ; i++ )
		pshape[i] = 1;

	if ( check_pshape( prank, pshape, ntasks ) )
		return;

	/* Factor in Nprocs */

	np = ntasks;

	i = 0;

	do
	{
		if ( i < prank - 1 )
		{
			factor = 2;

			while ( !STV_DIVISIBLE( np, factor ) )
				factor++;
		
			pshape[i] = factor;

			np /= factor;
		}

		else
			pshape[i] = np;

		i++;
	}
	while ( !check_pshape( prank, pshape, ntasks ) );
}


check_pshape( prank, pshape, ntasks )
int prank;
int *pshape;
int ntasks;
{
	int tot;
	int i;

	tot = 1;

	for ( i=0 ; i < prank ; i++ )
		tot *= pshape[i];

	if ( tot == ntasks )
		return( stvTrue );
	
	else
		return( stvFalse );
}


void
check_decomp_info( D )
struct decomp_info *D;
{
	int pindex;
	int index;
	int size;
	int part;
	int np;
	int p;

	pindex = 0;

	for ( index=0 ; index < D->data_dim ; index++ )
	{
		switch ( D->axis_type[ index ] )
		{
			case stvBlock:
			{
				size = DIM_SIZE( D, index );

				np = (int) ceil( (double) size
					/ (double) D->axis_info1[ index ] );

				if ( np > D->pshape[ pindex ] )
				{
					D->axis_info1[ index ] = (int) ceil( (double) size
						/ (double) D->pshape[ pindex ] );
				}

				pindex++;

				break;
			}

			case stvCyclic:
			{
				pindex++;
				
				break;
			}

			case stvExplicit:
			{
				part = (int) ceil( ((double) DIM_SIZE( D, index ))
					/ ((double) D->pshape[ pindex ]) );

				p = myinst % D->pshape[ pindex ];

				D->axis_bounds[ index ][0] =
					( p * part ) + D->glb[ index ];

				D->axis_bounds2[ index ][0] =
					( ( ( p + 1 ) * part ) - 1 ) + D->glb[ index ];

				D->axis_info1[ index ] = D->axis_bounds[ index ][0];
				D->axis_info2[ index ] = D->axis_bounds2[ index ][0];

				pindex++;

				break;
			}

			case stvCollapse: break;

			default:
			{
				printf(
					"Error:  Invalid Decomposition Axis Type = %d\n",
					D->axis_type[ index ] );
				fflush( stdout );

				break;
			}
		}
	}
}


void
set_paddr( prank, pshape, p, paddr )
int prank;
int *pshape;
int p;
int *paddr;
{
	int size;
	int i;

	size = 1;

	for ( i=(prank - 1) ; i >= 0 ; i-- )
	{
		paddr[i] = ( p / size ) % pshape[i];

		size *= pshape[i];
	}
}


get_var_alloc_size( D, paddr, array_decl )
struct decomp_info *D;
int *paddr;
int *array_decl;
{
	int i, j;
	int num;

	num = 1;

	j = 0;

	for ( i=0 ; i < D->data_dim ; i++ )
	{
		array_decl[i] = stv_get_proc_dim_size(
			D->axis_type[i], D->axis_info1[i],
			D->axis_bounds[i], D->axis_bounds2[i], D->nbounds[i],
			DIM_SIZE( D, i ), paddr[j], D->pshape[j] );

		if ( array_decl[i] < 0 )
		{
			printf( "stv_get_proc_dim_size() failed.\n" );

			return( -1 );
		}

		num *= array_decl[i];

		if ( D->axis_type[i] != stvCollapse )
			j++;
	}

	return( num );
}


void
fill_var_array( varptr, D, array_decl, type, paddr )
STV_VALUE varptr;
struct decomp_info *D;
int *array_decl;
int type;
int *paddr;
{
	STV_VALUE ptr;

	int indices[STV_MAX_DIM];
	int max[STV_MAX_DIM];

	int index;
	int i, j;
	int val;

	/* Initialize Indicies & Maximums */

	j = 0;

	for ( i=0 ; i < D->data_dim ; i++ )
	{
		max[i] = DIM_SIZE( D, i );

		switch ( D->axis_type[i] )
		{
			case stvBlock:
			{
				indices[i] = paddr[j] * D->axis_info1[i];

				j++;

				break;
			}

			case stvCyclic:
			{
				indices[i] = paddr[j] * D->axis_info1[i];

				j++;

				break;
			}

			case stvExplicit:
			{
				indices[i] = D->axis_info1[i] - D->glb[i];

				j++;

				break;
			}

			case stvCollapse:
			{
				indices[i] = 0;

				break;
			}

			default:
			{
				printf( "Error: Unknown Decomposition Type = %d\n",
					D->axis_type[i] );
				fflush( stdout );

				return;
			}
		}
	}

	/* Fill Local Array */

	ptr = varptr;

	do
	{
		val = val_of_index( D, indices );

		switch ( type )
		{
			case stvInt:
			{
				STV_VALUE_OF( ptr, int ) = (int) val;

				ptr = (STV_VALUE) ( STV_PTR_OF( ptr, int ) + 1 );

				break;
			}

			case stvFloat:
			{
				STV_VALUE_OF( ptr, float ) = (float) val;

				ptr = (STV_VALUE) ( STV_PTR_OF( ptr, float ) + 1 );

				break;
			}

			case stvDouble:
			{
				STV_VALUE_OF( ptr, double ) = (double) val;

				ptr = (STV_VALUE) ( STV_PTR_OF( ptr, double ) + 1 );

				break;
			}
		}
	}
	while ( incr_decomp_index( D, indices, array_decl, max ) );
}


val_of_index( D, indices )
struct decomp_info *D;
int *indices;
{
	int val;
	int sz;
	int i;

	val = 0;

	sz = 1;

	for ( i=0 ; i < D->data_dim ; i++ )
	{
		val += indices[i] * sz;

		sz *= DIM_SIZE( D, i );
	}

	return( val );
}


incr_decomp_index( D, indices, array_decl, max )
struct decomp_info *D;
int *indices;
int *array_decl;
int *max;
{
	int blk_chk;
	int pindex;
	int index;
	int flag;

	pindex = 0;

	index = 0;

	do
	{
		flag = 0;

		(indices[index])++;

		switch ( D->axis_type[index] )
		{
			case stvBlock:
			{
				blk_chk = indices[index] % D->axis_info1[index];

				if ( !blk_chk || blk_chk == array_decl[index] )
				{
					/* Reset Current Index */

					indices[index] -= array_decl[index];

					/* Bump Up Next Index */

					flag++;
				}

				break;
			}

			case stvCyclic:
			{
				if ( STV_DIVISIBLE( indices[index],
						D->axis_info1[index] ) )
				{
					/* Advance Current Index */

					indices[index] += D->axis_info1[index]
						* ( D->pshape[pindex] - 1 );

					/* Check for End of Dimension */

					if ( indices[index] >= max[index] )
					{
						/* Roll Back to First Cycle */

						indices[index] %= D->axis_info1[index]
							* D->pshape[pindex];

						/* Bump Up Next Index */

						flag++;
					}
				}

				break;
			}

			case stvExplicit:
			{
				if ( indices[index] >=
					D->axis_info2[index] - D->glb[index] + 1 )
				{
					indices[index] = D->axis_info1[index]	
						- D->glb[index];

					flag++;
				}

				break;
			}

			case stvCollapse:
			{
				if ( indices[index] >= array_decl[index] )
				{
					indices[index] = 0;

					flag++;
				}

				break;
			}
		}

		if ( indices[index] < 0 || indices[index] > max[index] )
		{
			printf( "Whoa!!!  indices[%d] = %d\n",
				index, indices[index] );
		}

		if ( flag )
		{
			if ( D->axis_type[index] != stvCollapse )
				pindex++;

			index++;

			if ( index >= D->data_dim )
				return( stvFalse );
		}
	}
	while ( flag );

	return( stvTrue );
}


PART
create_part()
{
	PART tmp;

	int i;

	tmp = (PART) malloc( sizeof( struct part_struct ) );

	if ( stv_memcheck( tmp, "User Particle Structure" ) != stvStatusOk )
		exit( -1 );

	for ( i=0 ; i < STV_MAX_DIM ; i++ )
		tmp->coords[i] = -1;

	tmp->foo_double = -1.0;

	for ( i=0 ; i < 2 ; i++ )
		tmp->foo_float[i] = -1.0;

	for ( i=0 ; i < 4 ; i++ )
		tmp->foo_int[i] = -1;

	tmp->next = (PART) NULL;

	return( tmp );
}


void
free_part( ptr )
PART *ptr;
{
	PART P;

	int i;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	P = *ptr;

	for ( i=0 ; i < STV_MAX_DIM ; i++ )
		P->coords[i] = -1;

	P->foo_double = -1.0;

	for ( i=0 ; i < 2 ; i++ )
		P->foo_float[i] = -1.0;

	for ( i=0 ; i < 4 ; i++ )
		P->foo_int[i] = -1;

	P->next = (PART) NULL;

	free( P );

	*ptr = (PART) NULL;
}


void
get_part( pnum, R, id, coords )
int pnum;
STV_REGION R;
STV_PARTICLE_ID *id;
int *coords;
{
	static PART P_save = (PART) NULL;
	static int count_save = -1;

	PART P;

	int count;

	/* Reset */

	if ( pnum == -1 )
	{
		*id = (STV_VALUE) NULL;

		P_save = (PART) NULL;

		count_save = -1;

		return;
	}

	/* Restore */

	if ( pnum >= count_save && count_save != -1 )
	{
		P = P_save;

		count = count_save;
	}

	/* Fresh Search */

	else
	{
		P = PART_LIST;

		count = 0;
	}

	/* Search for Particle */

	while ( P != NULL )
	{
		if ( stv_particle_in_region( R, P->coords, 3 ) )
		{
			count++;

			if ( count == pnum )
			{
				*id = (STV_VALUE) P;

				coords[0] = P->coords[0];
				coords[1] = P->coords[1];
				coords[2] = P->coords[2];

				P_save = P->next;

				count_save = count;

				return;
			}
		}

		P = P->next;
	}

	/* Particle Not Found */

	*id = (STV_VALUE) NULL;

	P_save = (PART) NULL;

	count_save = -1;
}


void get_foo_pfield( id, field, value )
STV_PARTICLE_ID id;
STV_VALUE field;
STV_VALUE *value;
{
	PART P;

	long fid;

	P = (PART) id;

	fid = (long) field;

	switch ( fid )
	{
		case PF_DOUBLE:
			*value = (STV_VALUE) &(P->foo_double);
			break;

		case PF_FLOAT:
			*value = (STV_VALUE) &(P->foo_float);
			break;

		case PF_INT:
			*value = (STV_VALUE) &(P->foo_int);
			break;

		default:
			*value = (STV_VALUE) NULL;
	}
}


void make_part( coords, id )
int *coords;
STV_PARTICLE_ID *id;
{
	PART P;

	if ( PART_LIST == NULL )
		P = PART_LIST = create_part();
	
	else
	{
		P = PART_LIST;

		while ( P->next != NULL )
			P = P->next;

		P = P->next = create_part();
	}

	P->coords[0] = coords[0];
	P->coords[1] = coords[1];
	P->coords[2] = coords[2];

	*id = (STV_VALUE) P;
}


void set_foo_pfield( id, field, value )
STV_PARTICLE_ID id;
STV_VALUE field;
STV_VALUE value;
{
	PART P;

	float *fptr;

	int *iptr;

	long fid;

	int i;

	P = (PART) id;

	fid = (long) field;

	switch ( fid )
	{
		case PF_DOUBLE:
			P->foo_double = STV_VALUE_OF( value, double );
			break;

		case PF_FLOAT:
			fptr = (float *) value;
			for ( i=0 ; i < 2 ; i++ )
				P->foo_float[i] = *fptr++;
			break;

		case PF_INT:
			iptr = (int *) value;
			for ( i=0 ; i < 4 ; i++ )
				P->foo_int[i] = *iptr++;
			break;

		default:
			break;
	}
}


void
read_args( argc, argv )
int argc;
char **argv;
{
	int i, j, k;

	APPNAME = (char *) NULL;

	NDECOMPS = 1;
	NPARAMS = 1;
	NFIELDS = 1;

	MSGTAG = 1;
	NTASKS = 1;
	NSPAWN = 1;

	SIZE = 32;

	SPMD = stvFalse;

	vflag = 0;
	dflag = 0;

	stv_setopt( stvOptNull, stvOptVerbose, vflag ? STV_DEBUG_ALL : 0 );

	for ( i=1 ; i < argc ; i++ )
	{
		if ( !strcmp( argv[i], "-SPMD" ) )
			SPMD = stvTrue;

		else if ( argv[i][0] == '-' )
		{
			k = i + 1;

			for ( j=1 ; j < strlen( argv[i] ) ; j++ )
			{
				switch ( argv[i][j] )
				{
					case 'v':
					{
						vflag++;

						stv_setopt( stvOptNull, stvOptVerbose,
							vflag ? STV_DEBUG_ALL : 0 );

						break;
					}

					case 'd':  dflag++;  break;

					case 'A':
					{
						if ( APPNAME != NULL )
							free( APPNAME );

						APPNAME = stv_copy_str( argv[k++] );

						break;
					}

					case 'S':
					{
						SIZE = atoi( argv[k++] );

						break;
					}

					case 'D':
					{
						NDECOMPS = atoi( argv[k++] );

						NDECOMPS = NDECOMPS > MAX_DECOMP ?
							MAX_DECOMP : NDECOMPS;

						break;
					}

					case 'P':
					{
						NPARAMS = atoi( argv[k++] );

						NPARAMS = NPARAMS > MAX_PARAM ?
							MAX_PARAM : NPARAMS;

						break;
					}

					case 'F':
					{
						NFIELDS = atoi( argv[k++] );

						NFIELDS = NFIELDS > MAX_FIELD ?
							MAX_FIELD : NFIELDS;

						break;
					}

					case 'M':
					{
						MSGTAG = atoi( argv[k++] );

						break;
					}

					case 'N':
					{
						NTASKS = atoi( argv[k++] );

						break;
					}

					case 'R':
					{
						NSPAWN = atoi( argv[k++] );

						break;
					}

					case 'h':
					{
						usage();

						break;
					}

					default:
					{
						usage();

						break;
					}
				}
			}

			i = k - 1;
		}
	}
}


void
usage()
{
	printf( "usage:  apptest [-vd] [-A <appname>] [-S <size>] " );
	printf( "[-D <ndecomp>] [-P <nparam>]\n" );
	printf( "\t[-F <nfield>] [-N <ntasks>] [-M <msgtag>] " );
	printf( "[-R <nspawn>] [-SPMD]\n" );

	printf( "\nwhere:\n" );

	printf( "\t-v = run in verbose mode, major output dumping.\n" );

	printf( "\t-d = spawn in debug mode (PvmTaskDebug).\n" );

	printf( "\t-A <appname> = name to post for application.\n" );

	printf( "\t-S <size> = order of matrix size in each dimension.\n" );

	printf( "\t-D <ndecomp> = # of decompositions to create.\n" );
	printf( "\t-P <nparam> = # of parameters to create.\n" );
	printf( "\t-F <nfield> = # of fields to create.\n" );

	printf( "\t-N <ntasks> = # of SPMD tasks being spawned.\n" );
	printf( "\t-M <msgtag> = message tag for STV communication.\n" );

	printf(
		"\t-R <nspawn> = # of recovery tasks to spawn on notify.\n" );

	printf( "\t-SPMD = do not use master to spawn rest of tasks.\n" );

	exit( 0 );
}


char **add_argv( argv, argc, newarg )
char **argv;
int argc;
char *newarg;
{
	char **tmp;

	int i;

	tmp = (char **) malloc( (STV_SIZE_T) (argc + 2) * sizeof(char *) );

	if ( stv_memcheck( tmp, "New Argv List Pointer" ) != stvStatusOk )
		exit( -1 );

	for ( i=0 ; i < argc ; i++ )
		tmp[i] = argv[i];

	tmp[argc] = newarg;

	tmp[argc+1] = (char *) NULL;

	return( tmp );
}


void
program_init()
{
	int i;

	/* Clear Field Variable Pointers */

	for ( i=0 ; i < MAX_FIELD ; i++ )
		varptrs[i] = (STV_VALUE) NULL;

	/* Clear Particle List */

	PART_LIST = (PART) NULL;

	NPARTS = 0;
}


void
clear_stv_stuff( field_ids, nfields, param_ids, part_ids )
int *field_ids;
int *nfields;
int *param_ids;
int *part_ids;
{
	PART Pnext;
	PART P;

	struct decomp_info *D;

	int i, j;

	/* Clear Particle IDs List */

	for ( i=0 ; i < MAX_PART ; i++ )
		part_ids[i] = -1;

	/* Clear Parameter IDs List */

	for ( i=0 ; i < MAX_PARAM ; i++ )
		param_ids[i] = -1;

	/* Clear Field IDs List */

	for ( i=0 ; i < MAX_FIELD ; i++ )
		field_ids[i] = -1;

	*nfields = -1;

	/* Clear Parameters */

	for ( i=0 ; i < MAX_PARAM ; i++ )
	{
		dparameters[i] = -1.0;

		fparameters[i] = -1.0;

		iparameters[i] = -1;
	}

	/* Clear Decompositions */

	for ( i=0 ; i < MAX_DECOMP ; i++ )
	{
		D = &(decomps[i]);

		D->id = -1;

		D->data_dim = -1;

		for ( j=0 ; j < STV_MAX_DIM ; j++ )
		{
			D->axis_type[j] = -1;

			D->axis_info1[j] = -1;

			D->axis_info2[j] = -1;

			D->glb[j] = -1;

			D->gub[j] = -1;

			D->pshape[j] = -1;
		}

		D->ntasks = -1;

		D->prank = -1;
	}

	/* Clear Fields */

	for ( i=0 ; i < MAX_FIELD ; i++ )
	{
		if ( varptrs[i] != NULL )
		{
			free( varptrs[i] );

			varptrs[i] = (STV_VALUE) NULL;
		}
	}

	/* Clear Particles */

	P = PART_LIST;

	while ( P != NULL )
	{
		Pnext = P->next;

		free_part( &P );

		P = Pnext;
	}

	NPARTS = 0;
}

