
/* $Id: viewer.c,v 3.37 2000/07/18 14:02:43 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>


/* STV Header Files */

#include "stv.h"


/* Commands String */

#define CMD_STR    "(y/#/v/r/f/U/s/R/V/d/D/C/F/L/l/u/o/O/n/N/q/?)"


/* XON Macro */

#define DO_XON( _vfg, _v ) \
\
	printf( "\nSending XON at " ); \
	stv_dump_time_str( stdout, (_vfg)->region_time, stvTrue ); \
	fflush( stdout ); \
\
	stv_viewer_send_XON( _vfg, _v ); \
\

#define DO_XON_CNT( _vfg, _v, _cnt ) \
\
	printf( "\nSending XON at " ); \
	stv_dump_time_str( stdout, (_vfg)->region_time, stvFalse ); \
	printf( " (%d)\n", xon_cnt ); \
	fflush( stdout ); \
\
	stv_viewer_send_XON( _vfg, _v ); \
\


/* Routines */

void	cleanup();
void	dump_region();
void	free_field_names();
void	handle_steering();
void	otf_usage();
void	program_init();
void	prompt_for_vis_region();
void	read_args();
void	read_field_names();
void	read_freq_synch_static();
void	read_name();
void	usage();


/* Globals */

STV_VIEWER VIEWER;

STV_REGION VISREGION;

STV_CELL VISCELL;

char **FIELDNAMES;
int NFIELDNAMES;

int STORAGE_ORDER;

int RECONNECT;

int STEERING;

int RESTART;

int MYTID;

int vflag;
int Dflag;
int dflag;


/* Main */

main( argc, argv )
int argc;
char **argv;
{
	/* Program Initialization */

	program_init();

	read_args( argc, argv );

	while ( 1 )
	{
		if ( do_viewer() )
			exit( 0 );

		if ( !RESTART )
			cleanup();
	}
}


do_viewer()
{
	STV_VIEW_FIELD_GROUP CURRENT_VFG;
	STV_VIEW_FIELD_GROUP VFG;

	STV_VIEW_FIELD VF_PRIMARY;
	STV_VIEW_FIELD VF;

	STV_PARTICLEPTR PT;

	STV_DECOMPPTR D;

	char appname[1024];
	char tmpname[1024];
	char prompt[1024];

	int option;
	int value;
	int what;

	int is_static;
	int no_synch;
	int whacked;
	int xon_cnt;
	int realloc;
	int newapp;
	int fflag;
	int i, j;
	int done;
	int freq;
	int quit;
	int cnt;
	int cc;

	/* Set Allocation Flag */

	realloc = stvTrue;

	/* Prompt for Viewer Info */

	fflag = 0;

	if ( !RESTART )
	{
		printf( "\nEnter Application Name / %s / %s: ",
			"<return> for unnamed", "\"quit\"" );

		read_name( appname );

		if ( !strcmp( appname, "quit" ) )
			return( stvTrue );

		read_field_names( &FIELDNAMES, &NFIELDNAMES, stvFalse );

		if ( NFIELDNAMES )
		{
			read_freq_synch_static( &freq, &no_synch, &is_static );

			fflag = 1;
		}
	}

	/* Initialize Viewer & Connect to Application */

	cnt = 0;

	do
	{
		cc = stv_viewer_init( &VIEWER, MYTID, appname, stvFalse );
	}
	while ( cc != stvStatusOk && ++cnt < 10 );

	if ( cc != stvStatusOk )
	{
		printf( "\nApplication \"%s\" Not Found - Reset.\n", appname );

		return( stvFalse );
	}

	/* Check for Field Selection */

	newapp = 0;

	quit = 0;

	do
	{
		if ( !RESTART )
		{
			do
			{
				while ( !NFIELDNAMES )
				{
					printf( "\nAvailable Fields are:\n" );
	
					VF = VIEWER->vfields;
	
					while ( VF != NULL )
					{
						if ( VF->field != NULL
							&& VF->field->name != NULL )
						{
							if ( VF->field->decomp != NULL )
							{
								D = VF->field->decomp;
	
								printf( "\t%s ", VF->field->name );
	
								for ( i=0 ; i < D->dataDim ; i++ )
								{
									printf( "[ %d - %d ]",
										D->glb[i], D->gub[i] );
								}
							}

							else if ( VF->field->particle != NULL )
							{
								PT = VF->field->particle;
	
								printf( "\t%s (%s) ", VF->field->name,
									PT->name );
	
								for ( i=0 ; i < PT->dataDim ; i++ )
								{
									printf( "[ %d - %d ]",
										PT->glb[i], PT->gub[i] );
								}
							}
	
							printf( "\n" );
						}
	
						VF = VF->next;
					}
	
					read_field_names( &FIELDNAMES, &NFIELDNAMES,
						stvTrue );
				}

				whacked = select_view_fields( &VF_PRIMARY,
					&FIELDNAMES, &NFIELDNAMES );

				if ( !whacked )
				{
					/* Read in Vis Region Fields */

					prompt_for_vis_region( VF_PRIMARY->field,
						VISREGION, VISCELL, stvTrue );

					dump_region( VF_PRIMARY->field,
						VISREGION, VISCELL );

					if ( !fflag )
						read_freq_synch_static( &freq, &no_synch,
							&is_static );
				}
			}
			while ( whacked );
		}

		else
		{
			RESTART = 0;

			if ( select_view_fields( &VF_PRIMARY,
				&FIELDNAMES, &NFIELDNAMES ) )
			{
				return( stvFalse );
			}
		}

		/* Request Field from Application */

		CURRENT_VFG = stv_viewer_submit_field_request( VIEWER,
			VISREGION, VISCELL, freq,
			( stvFieldRqstDefault
				| ( ( no_synch ) ? stvFieldRqstNoSynch : 0 )
				| ( ( is_static ) ? stvFieldRqstStatic : 0 ) ),
			&RESTART );

		if ( CURRENT_VFG == NULL )
			return( stvFalse );

		/* Ready to Receive Frames */

		printf( "\nReady to Receive Frames\n\n" );
		fflush( stdout );

		xon_cnt = 0;

		done = 0;

		do
		{
			/* Allocate Initial Field Storage */

			if ( realloc )
			{
				for ( i=0 ; i < CURRENT_VFG->nvfields ; i++ )
				{
					VF = CURRENT_VFG->view_field_list[i];

					if ( stv_allocate_Field_arrayDecl( CURRENT_VFG, i )
						!= stvStatusOk )
					{
						return( stvFalse );
					}

					if ( VF->field->decomp != NULL )
					{
						if ( stv_allocate_Field_Data( CURRENT_VFG, i,
								CURRENT_VFG->view_types[i] )
							!= stvStatusOk )
						{
							return( stvFalse );
						}

						if ( Dflag )
						{
							printf(
						"\nInitial Global \"%s\" Field Frame Data:\n",
								VF->field->name );
							fflush( stdout );

							stv_dump_vfg_field_array( CURRENT_VFG, i );
						}
					}
				}

				realloc = stvFalse;
			}

			/* Collect Data Frame from Application */

			do
			{
				cc = stv_viewer_receive_frame( &VFG, VIEWER, &RESTART,
					stvMsgBlock );

				if ( cc != stvStatusOk && cc != stvStatusBadFrame )
				{
					/* Free Data Frame */

					for ( i=0 ; i < CURRENT_VFG->nvfields ; i++ )
						stv_free_field_array( CURRENT_VFG, i );

					stv_reset_Field_Particles( CURRENT_VFG );

					/* Release Application */

					printf( "\nSending FieldHaltGroup\n" );
					fflush( stdout );

					stv_viewer_send_FieldHaltGroup( CURRENT_VFG,
						VIEWER );

					/* Handle Re-Connect */

					if ( RECONNECT )
						RESTART++;

					return( stvFalse );
				}

				if ( VFG != CURRENT_VFG )
					stv_viewer_release_field( &VFG, VIEWER );
			}
			while ( VFG != CURRENT_VFG );

			/**** Use Data Frame Here ****/

			/* Free Particle Field Data */

			stv_reset_Field_Particles( CURRENT_VFG );

			/* Check App Lock (& Reset) */

			if ( CURRENT_VFG->locked == stvAppLocked && xon_cnt )
			{
				printf(
					"Application Locked!  Resetting XON Count...\n" );
				xon_cnt = 0;
			}

			/* Handle Viewer XON Iterations */

			if ( xon_cnt )
			{
				DO_XON_CNT( CURRENT_VFG, VIEWER, xon_cnt );

				xon_cnt--;
			}

			/* Prompt Viewer for Action */

			else
			{
				do
				{
					printf( "Send XON Message %s: ", CMD_STR );

					scanf( "%s", prompt );

					if ( prompt[0] == '?' )
					{
						otf_usage();

						*prompt = '\0';
					}
				}
				while ( !strcmp( prompt, "" ) );

				/* Grab '\n' */

				while ( getchar() != '\n' );

				/* Send Multiple XON Messages */

				if ( sscanf( prompt, "%d", &xon_cnt ) == 1 )
				{
					DO_XON_CNT( CURRENT_VFG, VIEWER, xon_cnt );

					xon_cnt--;
				}

				/* Send XON Message */

				else if ( prompt[0] == 'y' )
				{
					DO_XON( CURRENT_VFG, VIEWER );
				}

				/* Read in New Vis Region */

				else if ( prompt[0] == 'v'
					|| prompt[0] == 'r' || prompt[0] == 'f' )
				{
					/* Read in New Vis Region Fields */

					if ( prompt[0] == 'v' || prompt[0] == 'r' )
					{
						/* Skip if View Field Group is Static */
						/* (return code handling included below */
						/* anyway as an example...) */

						if ( is_static )
						{
							printf( "Static View Field Group - " );
							printf( "Can't Change View Region!\n" );
						}

						else
						{
							prompt_for_vis_region( VF_PRIMARY->field,
								VISREGION, VISCELL, stvFalse );

							dump_region( VF_PRIMARY->field,
								VISREGION, VISCELL );

							printf( "\nSending Vis Region\n" );
							fflush( stdout );

							if ( stv_viewer_set_VisRegion( CURRENT_VFG,
									VIEWER, VISREGION, VISCELL )
								!= stvStatusOk )
							{
								printf(
						"Couldn't Change View Region - Reverting.\n" );

								stv_copy_region( VISREGION,
									CURRENT_VFG->visregion );
							
								stv_copy_cell( VISCELL,
									CURRENT_VFG->viscell );
							}

							else
								realloc = stvTrue;
						}
					}

					if ( prompt[0] == 'v' || prompt[0] == 'f' )
					{
						read_freq_synch_static( &freq,
							(int *) NULL, (int *) NULL );

						printf( "\nSending Vis Frequency\n" );
						fflush( stdout );

						stv_viewer_set_VisFrequency( CURRENT_VFG,
							VIEWER, freq );
					}

					/* Send XON */

					DO_XON( CURRENT_VFG, VIEWER );
				}

				/* Toggle Full Unpack Option */

				else if ( prompt[0] == 'U' )
				{
					cc = stv_setopt( stvOptNull, stvOptFullUnpack,
						!stv_getopt( stvOptNull, stvOptFullUnpack ) );

					printf( "\nTurned Full Unpack %s\n\n",
						cc ? "OFF" : "ON" );
					fflush( stdout );

					DO_XON( CURRENT_VFG, VIEWER );

					realloc = stvTrue;
				}

				/* Request Steering */

				else if ( prompt[0] == 's' )
				{
					handle_steering( freq, no_synch );

					DO_XON( CURRENT_VFG, VIEWER );
				}

				/* Toggle Verbosity Option */

				else if ( prompt[0] == 'V' )
				{
					printf( "\nEnter Verbosity Level %s:",
							"(or <return> for STV_DEBUG_ALL)" );
					fflush( stdout );

					if ( scanf( "%d", &value ) != 1 )
						value = STV_DEBUG_ALL;

					stv_setopt( stvOptNull, stvOptVerbose, value );

					vflag = value ? stvTrue : stvFalse;

					printf( "\nVerbosity Set to 0x%x.\n\n", value );
					fflush( stdout );

					DO_XON( CURRENT_VFG, VIEWER );
				}

				/* Toggle Dump Verbosity Option */

				else if ( prompt[0] == 'D' )
				{
					Dflag = Dflag ? stvFalse : stvTrue;

					stv_setopt( stvOptNull, stvOptDumpAll, Dflag );

					printf( "\nTurned Updated Field Dumping %s\n\n",
						Dflag ? "ON" : "OFF" );
					fflush( stdout );

					DO_XON( CURRENT_VFG, VIEWER );
				}

				/* Toggle Final Dump Option */

				else if ( prompt[0] == 'd' )
				{
					dflag = dflag ? stvFalse : stvTrue;

					stv_setopt( stvOptNull, stvOptDump, dflag );

					printf( "\nTurned Final Field Dumping %s\n\n",
						dflag ? "ON" : "OFF" );
					fflush( stdout );

					DO_XON( CURRENT_VFG, VIEWER );
				}

				/* Lock Application */

				else if ( prompt[0] == 'L' )
				{
					printf( "\nLocking Whole Application...\n" );
					fflush( stdout );

					stv_viewer_send_AppLock(
							(STV_VIEW_FIELD_GROUP) NULL, VIEWER, 0 );

					DO_XON( CURRENT_VFG, VIEWER );
				}

				/* Lock Application */

				else if ( prompt[0] == 'l' )
				{
					printf( "\nLocking Application Group...\n" );
					fflush( stdout );

					stv_viewer_send_AppLock( CURRENT_VFG, VIEWER, 0 );

					DO_XON( CURRENT_VFG, VIEWER );
				}

				/* Unlock Application */

				else if ( prompt[0] == 'u' )
				{
					printf( "\nUnlocking Application...\n" );
					fflush( stdout );

					stv_viewer_send_AppUnlock( CURRENT_VFG, VIEWER );

					DO_XON( CURRENT_VFG, VIEWER );
				}

				/* Set Application's CUMULVS Options On-The-Fly */
				/* (That can't be a problem, can it? :-) */

				else if ( prompt[0] == 'o' || prompt[0] == 'O' )
				{
					printf( "\nSend CUMULVS Options to %s.\n\n",
						( prompt[0] == 'o' )
							? "Current Application Group"
							: "Entire Application" );

					printf( "'What' Constants:\n" );
					printf( "\t%d = %s\n",
						stvOptAll, stv_option_str( stvOptAll ) );
					printf( "\t%d = %s\n",
						stvOptNull, stv_option_str( stvOptNull ) );

					printf( "\nEnter 'what' for option: " );
					fflush( stdout );

					if ( scanf( "%d", &what ) != 1 )
						what = stvOptNull;

					printf( "Options:\n" );

					for ( i=0 ; i < stvOptMax ; i++ )
						printf( "\t%d = %s\n", i, stv_option_str( i ) );
					
					printf( "\nEnter option: " );
					fflush( stdout );

					if ( scanf( "%d", &option ) != 1 )
						option = stvOptNull;

					printf( "\nEnter option value: " );
					fflush( stdout );

					if ( scanf( "%d", &value ) != 1 )
						value = stvOptNull;

					if ( prompt[0] == 'o' )
					{
						stv_viewer_send_Setopt( CURRENT_VFG, VIEWER,
								what, option, value );
					}
					
					else
					{
						stv_viewer_send_Setopt(
								(STV_VIEW_FIELD_GROUP) NULL, VIEWER,
								what, option, value );
					}

					DO_XON( CURRENT_VFG, VIEWER );
				}

				/* Set Storage Order to C, Row Major */

				else if ( prompt[0] == 'C' )
				{
					STORAGE_ORDER = stvRowMajor;

					printf( "\nData Storage Set to Row Major\n\n" );

					printf( "\nSending FieldHaltGroup\n" );
					fflush( stdout );

					stv_viewer_send_FieldHaltGroup( CURRENT_VFG,
						VIEWER );

					realloc = stvTrue;

					RESTART++;

					done++;
				}

				/* Set Storage Order to Fortran, Column Major */

				else if ( prompt[0] == 'F' )
				{
					STORAGE_ORDER = stvColumnMajor;

					printf( "\nData Storage Set to Column Major\n\n" );

					printf( "\nSending FieldHaltGroup\n" );
					fflush( stdout );

					stv_viewer_send_FieldHaltGroup( CURRENT_VFG,
						VIEWER );

					realloc = stvTrue;

					RESTART++;

					done++;
				}

				/* Send FieldHalt Messages */

				else if ( prompt[0] == 'R' || prompt[0] == 'n'
					|| prompt[0] == 'N' || prompt[0] == 'q' )
				{
					printf( "\nSending FieldHaltGroup\n" );
					fflush( stdout );

					stv_viewer_send_FieldHaltGroup( CURRENT_VFG,
						VIEWER );

					realloc = stvTrue;

					done++;

					if ( prompt[0] == 'R' )
						RESTART++;

					if ( prompt[0] == 'N' )
						{ newapp++; quit++; }

					if ( prompt[0] == 'q' )
						quit++;
				}

				/* Command Help */

				else
				{
					printf( "\nUnknown Command '%c'.\n\n", prompt[0] );

					otf_usage();

					printf( "Hit Return to Continue: " );
					fflush( stdout );

					/* Grab '\n' */

					while ( getchar() != '\n' );

					/* Send XON & Continue */

					DO_XON( CURRENT_VFG, VIEWER );
				}
			}
		}
		while ( !done );

		if ( !RESTART )
			free_field_names( &FIELDNAMES, &NFIELDNAMES );
	}
	while ( !quit );

	if ( newapp )
		return( stvFalse );

	else
		return( stvTrue );
}


select_view_fields( VF_PRIMARY, fieldnames, nfields )
STV_VIEW_FIELD *VF_PRIMARY;
char ***fieldnames;
int *nfields;
{
	STV_VIEW_FIELD VF;

	int whacked;
	int i;

	stv_clear_view_field_select( VIEWER );

	whacked = stvFalse;

	for ( i=0 ; i < *nfields && !whacked ; i++ )
	{
		VF = stv_get_view_field_name( (*fieldnames)[i],
			VIEWER->vfields, VIEWER->nvfields );
	
		if ( VF == NULL )
		{
			printf( "Error: Field \"%s\" Not Found.\n",
				(*fieldnames)[i] );
	
			free_field_names( fieldnames, nfields );

			whacked++;
		}
	
		else
		{
			/* Snag Primary View Field */

			if ( !i )
				*VF_PRIMARY = VF;

			/* Select View Field */

			VF->selected = stvTrue;

			VF->view_type = VF->field->type;

			VF->view_storage_order =
				( ( STORAGE_ORDER >= 0 ) ?
					STORAGE_ORDER
					: VF->field->storage_order );
		}
	}

	return( whacked );
}


void
otf_usage()
{
	printf( "\nViewer Commands:\n\n" );

	printf( "y  -  " );
	printf( "Send XON to Application.\n" );

	printf( "#  -  " );
	printf( "Number of XON's to Send.\n" );

	printf( "v  -  " );
	printf( "Enter New Vis Region & Frame Freq.\n" );

	printf( "r  -  " );
	printf( "Enter New Visualization Region.\n" );

	printf( "f  -  " );
	printf( "Enter New Vis Frame Frequency.\n" );

	printf( "U  -  " );
	printf( "Toggle \"Full Unpack\" Option.\n" );

	printf( "s  -  " );
	printf( "Request Steering of Parameters.\n" );

	printf( "R  -  " );
	printf( "Initiate Restart of Application Link.\n" );

	printf( "V  -  " );
	printf( "Toggle Verbosity Option.\n" );

	printf( "d  -  " );
	printf( "Toggle Final Data Frame Dump Option.\n" );

	printf( "D  -  " );
	printf( "Toggle Mega Data Dumping Option.\n" );

	printf( "C  -  " );
	printf( "Force Data Storage Order to " );
	printf( "Row Major, a la C.\n" );

	printf( "F  -  " );
	printf( "Force Data Storage Order to " );
	printf( "Column Major, a la Fortran.\n" );

	printf( "L  -  " );
	printf( "Lock Whole Application at Next SendToFE Call.\n" );

	printf( "l  -  " );
	printf( "Lock Application Group at Next SendToFE Call.\n" );

	printf( "u  -  " );
	printf( "Unlock Application, Release to Continue " );
	printf( "Computation.\n" );

	printf( "o  -  " );
	printf( "Send CUMULVS Options to Current Application Group.\n" );

	printf( "O  -  " );
	printf( "Send CUMULVS Options to Entire Application.\n" );

	printf( "n  -  " );
	printf( "Do Not Send XON, Release Data Field.\n" );

	printf( "N  -  " );
	printf( "Do Not Send XON, Release Data Field, " );
	printf( "Select New Application.\n" );

	printf( "q  -  " );
	printf( "Quit Viewer, Disconnect Application.\n" );

	printf( "?  -  " );
	printf( "Generate This Command Help.\n" );

	printf( "\nNote:  All Commands Send XON Unless " );
	printf( "Specified Otherwise.\n\n" );
}


void
handle_steering( freq, no_synch )
int freq;
int no_synch;
{
	STV_VIEW_PARAM VP;

	STV_PARAMPTR P;

	double dval;

	char tmp[1024];

	char *ptr;

	int newsteer;
	int changed;
	int nsteer;
	int rcount;
	int count;
	int cc;
	int i;

	/* Initialize Steered Param Count */

	newsteer = 0;

	nsteer = 0;

	/* Prompt for Requests */

	printf( "\nSteering Parameter Request List:\n" );

	VP = VIEWER->vparams;

	rcount = 0;

	count = 0;

	while ( VP != NULL )
	{
		if ( VP->token == stvSteerNot )
		{
			printf( "\t%s\n", VP->param->name );

			count++;
		}

		else
			rcount++;

		VP = VP->next;
	}

	if ( count )
	{
		printf( "\nRequest New Parameter (or <return>)? " );

		read_name( tmp );

		while ( count && strcmp( tmp, "" ) )
		{
			VP = stv_get_view_param_name( tmp,
				VIEWER->vparams, VIEWER->nvparams );

			if ( VP == NULL )
			{
				printf( "\nError: Parameter \"%s\" Not Found.\n\n",
					tmp );
			}
		
			else
			{
				if ( !STEERING )
				{
					printf( "Initializing Steering\n" );

					cc = stv_viewer_steering_init( VIEWER,
						freq, no_synch, &RESTART );

					printf( "\tcc=%d\n", cc );

					if ( cc != stvStatusOk )
						return;

					STEERING = stvTrue;
				}

				stv_viewer_steering_request( VP, VIEWER,
						stvMsgNoBlock );

				if ( VP->token == stvSteerToken )
					newsteer++;

				else if ( VP->token == stvSteerRqstd )
				{
					printf(
						"* Steering Token for %s Not Yet Available *\n",
						VP->param->name );
				}

				count--;
			}

			if ( count )
			{
				printf(
					"\nRequest Another New Parameter (or <return>)? " );

				read_name( tmp );
			}
		}
	}

	else
		printf( "\t<empty list>\n" );

	/* Prompt for Releases */

	if ( rcount )
	{
		printf( "\nSteering Parameter Release List:\n" );

		VP = VIEWER->vparams;

		count = 0;

		while ( VP != NULL )
		{
			if ( VP->token != stvSteerNot )
			{
				printf( "\t%s\n", VP->param->name );

				if ( VP->token == stvSteerToken )
					nsteer++;

				else
				{
					stv_viewer_steering_request( VP, VIEWER,
							stvMsgNoBlock );

					if ( VP->token == stvSteerToken )
						nsteer++;

					else if ( VP->token == stvSteerRqstd )
					{
						printf(
						"* Steering Token for %s Not Yet Available *\n",
							VP->param->name );
					}
				}

				count++;
			}

			VP = VP->next;
		}

		if ( count )
		{
			printf( "\nRelease a Steering Parameter (or <return>)? " );

			read_name( tmp );

			while ( count && strcmp( tmp, "" ) )
			{
				VP = stv_get_view_param_name( tmp,
					VIEWER->vparams, VIEWER->nvparams );

				if ( VP == NULL )
				{
					printf( "\nError: Parameter \"%s\" Not Found.\n\n",
						tmp );
				}
		
				else
				{
					if ( !STEERING )
						printf( "\nWhoa! Not Steering, Release?!\n\n" );

					if ( VP->token == stvSteerToken )
						nsteer--;

					stv_viewer_steering_release( VP, VIEWER );

					count--;
				}

				if ( count )
				{
					printf(
			"\nRelease Another Steering Parameter (or <return>)? " );

					read_name( tmp );
				}
			}
		}

		else
			printf( "\t<empty list>\n" );
	}

	else
		nsteer += newsteer;

	/* Prompt for Actual Steering */

	if ( nsteer )
	{
		printf(
"\nAdjust the Values of the Steering Parameters (or <return>):\n\n" );

		VP = VIEWER->vparams;

		changed = 0;

		while ( VP != NULL )
		{
			VP->changed = stvFalse;

			if ( VP->token == stvSteerToken )
			{
				P = VP->param;

				for ( i=0 ; i < P->num ; i++ )
				{
					if ( P->num == 1 )
						printf( "%s ", P->name );

					else
						printf( "%s:%s ", P->name, P->pnames[i] );

					ptr = stv_str_of_typed_data( P->vars[i],
						P->types[i] & stvTypeMask );

					printf( "[%s]: ", ptr );

					free( ptr );

					read_name( tmp );

					if ( strcmp( tmp, "" ) )
					{
						if ( sscanf( tmp, "%lf", &dval ) == 1 )
						{
							if ( stv_viewer_steer_vparameter( VP,
								(STV_VALUE) &dval, stvDouble, i )
								!= stvStatusOk )
							{
								printf( "Error steering parameter!\n" );
							}

							else
								changed++;
						}

						else
						{
							printf( "\nError: Value Missing \"%s\"\n\n",
								tmp );
						}
					}

					else
						VP->changed = stvTrue;
				}
			}

			VP = VP->next;
		}

		if ( changed )
			stv_viewer_send_NewParams( VIEWER );
	}
}


void
read_field_names( fieldnames, nfields, known )
char ***fieldnames;
int *nfields;
int known;
{
	char tmp[1024];

	int found = 0;
	int i, j;
	int flag;
	int num;

	/* Only One Field to Choose From?  Now Let's See... */

	if ( VIEWER != NULL
			&& VIEWER->nvfields == 1
			&& VIEWER->vfields->field != NULL
			&& VIEWER->vfields->field->name != NULL )
	{
		*fieldnames = (char **) malloc( sizeof( char * ) );

		if ( stv_memcheck( *fieldnames, "Field Name List Ptr" )
			!= stvStatusOk )
		{
			exit( -1 );
		}

		printf( "\nOnly One Field to Choose From - " );
		printf( "Auto Select \"%s\".\n",
			VIEWER->vfields->field->name );

		(*fieldnames)[0] = stv_copy_str( VIEWER->vfields->field->name );

		*nfields = 1;

		return;
	}

	printf( "\nEnter Number of Fields / Single Field Name%s: ",
		( known ) ? "" : " / (<return> If Not Known)" );

	read_name( tmp );

	flag = sscanf( tmp, "%d", &num );

	if ( flag == 1 )
	{
		*fieldnames = (char **) malloc( (STV_SIZE_T) num
			* sizeof( char * ) );

		if ( stv_memcheck( *fieldnames, "Field Name List Ptr" )
			!= stvStatusOk )
		{
			exit( -1 );
		}

		for ( i=0 ; i < num ; i++ )
		{
			do
			{
				printf(
					"Enter Field Name #%d / (or <return> to reset): ",
					i + 1 );

				read_name( tmp );

				if ( !strcmp( tmp, "" ) )
				{
					*nfields = i;

					free_field_names( fieldnames, nfields );

					return;
				}

				else
					(*fieldnames)[i] = stv_copy_str( tmp );

				if ( known )
				{
					found = 0;

					if ( stv_get_view_field_name( (*fieldnames)[i],
						VIEWER->vfields, VIEWER->nvfields ) != NULL )
					{
						found++;
					}

					else
					{
						printf(
							"\nError: Field \"%s\" Not Found - %s.\n\n",
							(*fieldnames)[i], "Please Re-Enter" );
					}
				}
			}
			while ( known && !found );
		}

		*nfields = num;
	}

	else
	{
		/* Handle Single Field Case */

		if ( strlen( tmp ) > 0 )
		{
			*fieldnames = (char **) malloc( sizeof( char * ) );

			if ( stv_memcheck( *fieldnames, "Field Name List Ptr" )
				!= stvStatusOk )
			{
				exit( -1 );
			}

			(*fieldnames)[0] = stv_copy_str( tmp );

			*nfields = 1;

			return;
		}

		*fieldnames = (char **) NULL;

		*nfields = 0;
	}
}


void
free_field_names( fieldnames, nfields )
char ***fieldnames;
int *nfields;
{
	int i;

	for ( i=0 ; i < *nfields ; i++ )
		free( (*fieldnames)[i] );

	if ( *fieldnames != NULL )
	{
		free( *fieldnames );

		*fieldnames = (char **) NULL;
	}

	*nfields = 0;
}


void
read_freq_synch_static( freq, no_synch, is_static )
int *freq;
int *no_synch;
int *is_static;
{
	STV_VIEW_FIELD VF;

	char unitstr[255];
	char tmp[1024];

	double dnum;

	int units;
	int flag;
	int any;
	int num;
	int i;

	if ( freq != NULL )
	{
		any = 0;

		for ( i=0 ; i < NFIELDNAMES && VIEWER != NULL ; i++ )
		{
			VF = stv_get_view_field_name( FIELDNAMES[i],
				VIEWER->vfields, VIEWER->nvfields );
	
			if ( VF != NULL
					&& stv_is_time_value( VF->field->time_scale ) )
			{
				if ( !(any++) )
					printf( "\nSelected Field Time Scales Are:\n" );
				
				printf( "\t%s:  %d %s\n", VF->field->name,
					stv_value_of_time( VF->field->time_scale ),
					stv_time_unit_str(
						stv_unit_of_time( VF->field->time_scale ) ) );
			}
		}

		printf( "\nEnter Visualization Frame Frequency (1): " );
		read_name( tmp );

		flag = sscanf( tmp, "%lf %s", &dnum, unitstr );

		if ( flag < 1 )
			num = 1;

		else if ( flag == 2 )
		{
			units = stv_time_unit_of( unitstr );

			if ( stv_is_time_value( units ) )
			{
				num = stv_time_of_double( dnum, units );

				printf( "Frequency Interpreted as %d %s.\n",
					stv_value_of_time( num ),
					stv_time_unit_str( stv_unit_of_time( num ) ) );
			}
		}

		else
			num = (int) dnum;

		*freq = num;
	}

	if ( no_synch != NULL )
	{
		printf( "\nEnter NO_SYNCH Value [0/1] (0): " );
		read_name( tmp );

		flag = sscanf( tmp, "%d", &num );

		if ( flag != 1 )
			num = 0;

		*no_synch = num;
	}

	if ( is_static != NULL )
	{
		printf( "\nEnter Is_Static Value [0/1] (0): " );
		read_name( tmp );

		flag = sscanf( tmp, "%d", &num );

		if ( flag != 1 )
			num = 0;

		*is_static = num;
	}
}


void
prompt_for_vis_region( F, R, C, initflag )
STV_FIELDPTR F;
STV_REGION R;
STV_CELL C;
int initflag;
{
	char tmp[1024];

	int *glb, *gub;

	int flag;
	int dim;
	int num;
	int i;

	/* Get glb, gub */

	if ( F->decomp != NULL )
	{
		glb = F->decomp->glb;
		gub = F->decomp->gub;

		dim = F->decomp->dataDim;
	}

	else if ( F->particle != NULL )
	{
		glb = F->particle->glb;
		gub = F->particle->gub;

		dim = F->particle->dataDim;
	}

	else
	{
		printf( "\nWarning: No Decomp or Particle for Field \"%s\"\n\n",
			F->name );
		fflush( stdout );

		return;
	}

	/* Read in Vis Region Fields */

	for ( i=0 ; i < dim ; i++ )
	{
		if ( initflag )
			STV_REGION_MIN( R, i ) = glb[i];

		printf( "\nEnter Dimension #%d Min [ %d - %d ] (%d): ",
			i, glb[i], gub[i], STV_REGION_MIN( R, i ) );
		read_name( tmp );

		flag = sscanf( tmp, "%d", &num );

		if ( flag == 1 )
			STV_REGION_MIN( R, i ) = num;

		if ( initflag )
			STV_REGION_MAX( R, i ) = gub[i];

		printf( "Enter Dimension #%d Max [ %d - %d ] (%d): ",
			i, glb[i], gub[i], STV_REGION_MAX( R, i ) );
		read_name( tmp );

		flag = sscanf( tmp, "%d", &num );

		if ( flag == 1 )
		{
			if ( num < STV_REGION_MIN( R, i ) )
				num = STV_REGION_MIN( R, i );

			STV_REGION_MAX( R, i ) = num;
		}

		if ( F->decomp != NULL )
		{
			if ( initflag )
				STV_CELL_OF_DIM( C, i ) = 1;

			printf( "Enter Dimension #%d Cell Size (%d): ",
				i, STV_CELL_OF_DIM( C, i ) );
			read_name( tmp );

			flag = sscanf( tmp, "%d", &num );

			if ( flag == 1 )
				STV_CELL_OF_DIM( C, i ) = num;
		}
	}

	if ( F->particle != NULL )
	{
		if ( initflag )
			STV_CELL_PARTICLE( C ) = 1;

		printf( "\nEnter Particle Cell Size (%d): ",
			STV_CELL_PARTICLE( C ) );
		read_name( tmp );

		flag = sscanf( tmp, "%d", &num );

		if ( flag == 1 )
			STV_CELL_PARTICLE( C ) = num;
	}

	/* Dump Out Region */

	printf( "\n" );

	for ( i=0 ; i < dim ; i++ )
	{
		if ( F->decomp != NULL )
		{
			printf( "[%d]:  min=%d max=%d cell=%d\n", i,
				STV_REGION_MIN( R, i ), STV_REGION_MAX( R, i ),
				STV_CELL_OF_DIM( C, i ) );
		}

		else if ( F->particle != NULL )
		{
			printf( "[%d]:  min=%d max=%d\n", i,
				STV_REGION_MIN( R, i ), STV_REGION_MAX( R, i ) );
		}
	}

	if ( F->particle != NULL )
		printf( "\tparticle cell = %d\n", STV_CELL_PARTICLE( C ) );

	printf( "\n" );
}


void
dump_region( F, R, C )
STV_FIELDPTR F;
STV_REGION R;
STV_CELL C;
{
	int dim;
	int i;

	/* Get Data Dim */

	if ( F->decomp != NULL )
		dim = F->decomp->dataDim;

	else if ( F->particle != NULL )
		dim = F->particle->dataDim;

	else
	{
		printf( "\nWarning: No Decomp or Particle for Field \"%s\"\n\n",
			F->name );
		fflush( stdout );

		return;
	}

	printf( "Vis Region:\n" );

	if ( F->decomp != NULL )
	{
		for ( i=0 ; i < dim ; i++ )
		{
			printf( "\t[%d to %d] by %d\n",
				STV_REGION_MIN( R, i ), STV_REGION_MAX( R, i ),
				STV_CELL_OF_DIM( C, i ) );
		}
	}

	else if ( F->particle != NULL )
	{
		for ( i=0 ; i < dim ; i++ )
		{
			printf( "\t[%d to %d]\n",
				STV_REGION_MIN( R, i ), STV_REGION_MAX( R, i ) );
		}

		printf( "\t\tby %d\n", STV_CELL_PARTICLE( C ) );
	}
}


void
program_init()
{
	VIEWER = (STV_VIEWER) NULL;

	stv_init_region( VISREGION );

	stv_init_cell( VISCELL );

	STEERING = stvFalse;

	NFIELDNAMES = 0;

	RECONNECT = stvFalse;

	RESTART = 0;

	stv_myid( &MYTID );

	if ( MYTID < 0 )
	{
		stv_mp_err( "Obtaining My Task ID" );

		exit( 0 );
	}

	else
	{
		printf( "Viewer Alive as TID=[ 0x%x / %d ]\n", MYTID, MYTID );
		fflush( stdout );
	}

	/* Set STV Options */

	stv_setopt( stvOptNull, stvOptFullUnpack, stvFalse );
}


void
read_args( argc, argv )
int argc;
char **argv;
{
	int i, j, k;

	STORAGE_ORDER = -1;

	vflag = 0;

	Dflag = 0;
	dflag = 1;

	stv_setopt( stvOptNull, stvOptVerbose, vflag ? STV_DEBUG_ALL : 0 );
	stv_setopt( stvOptNull, stvOptDumpAll, Dflag );
	stv_setopt( stvOptNull, stvOptDump, dflag );

	for ( i=1 ; i < argc ; i++ )
	{
		if ( argv[i][0] == '-' )
		{
			k = i + 1;

			for ( j=1 ; j < strlen( argv[i] ) ; j++ )
			{
				switch ( argv[i][j] )
				{
					case 'v':
					{
						vflag++;
						
						stv_setopt( stvOptNull, stvOptVerbose,
							vflag ? STV_DEBUG_ALL : 0 );

						break;
					}

					case 'D':
					{
						Dflag++;
						
						stv_setopt( stvOptNull, stvOptDumpAll, Dflag );

						break;
					}

					case 'd':
					{
						dflag++;
						
						stv_setopt( stvOptNull, stvOptDump, dflag );

						break;
					}

					case 'C':
					{
						STORAGE_ORDER = stvRowMajor;

						break;
					}

					case 'F':
					{
						STORAGE_ORDER = stvColumnMajor;

						break;
					}

					case 'R':
					{
						RECONNECT = stvTrue;

						break;
					}

					case 'h':
					{
						usage();

						break;
					}

					default:
					{
						usage();

						break;
					}
				}
			}

			i = k - 1;
		}
	}
}


void
usage()
{
	printf( "usage:  viewer [-vDdCFR]\n" );

	printf( "\nwhere:\n" );

	printf( "\t-v = run in verbose mode.\n" );
	printf( "\t-D = major update dumping on.\n" );
	printf( "\t-d = final data dumping on.\n" );
	printf( "\t-C = use C data storage order, Row Major.\n" );
	printf( "\t-F = use Fortran data storage order, Column Major.\n" );
	printf( "\t-R = Reconnect on Application Failure / Recovery.\n" );

	printf( "\n" );

	exit( 0 );
}


void
cleanup()
{
	/* Reset STV Globals */

	stv_viewer_reset( VIEWER );

	stv_init_region( VISREGION );

	stv_init_cell( VISCELL );
}


void
read_name( str )
char *str;
{
	while ( (*str = getchar()) != '\n' )
		str++;

	*str = '\0';
}

