
/* $Id: stv_mp.h,v 3.26 1999/11/03 01:42:53 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


#ifndef _STV_MP_H_

#define _STV_MP_H_


/* Message passing macros / definitions */


#ifdef STV_MP_PVM


/* Message Passing Substrate Name */

#define STV_MP_NAME "PVM"


/* PVM Header Files */

#include "pvm3.h"
#include "pvmtev.h"


/* Context Type */

#define STV_CONTEXT			int

#define STV_CONTEXT_MATCH( _ctx1, _ctx2 ) \
	( (_ctx1) == (_ctx2) )

#define STV_CONTEXT_FPRINTF( _fp, _ctx ) \
	( fprintf( _fp, "%d", _ctx ) )

#define STV_CONTEXT_SPRINTF( _str, _ctx ) \
	sprintf( _str, "%d", _ctx ) 

#define STV_CONTEXT_SSCANF( _str, _ctxp ) \
	sscanf( _str, "%d", _ctxp )

#if ( PVM_MAJOR_VERSION > 3 ) || ( PVM_MINOR_VERSION >= 4 )
#define STV_BASECTX			PvmBaseContext
#else
#define STV_BASECTX			stvTrue
#endif


/* PVM Message Type Constants */

#define stvDataDefault		PvmDataDefault
#define stvDataRaw			PvmDataRaw
#define stvDataInPlace		PvmDataInPlace
#define stvTaskHost     	PvmTaskHost
#define stvTaskDebug		PvmTaskDebug
#define stvMppFront			PvmMppFront


/* PVM Notify Type Constants */

#define stvTaskExit			PvmTaskExit
#define stvHostDelete		PvmHostDelete
#define stvHostAdd			PvmHostAdd


/* Macro Temporary Storage */

extern	STV_CONTEXT STV_TMP_CTX;

extern	int STV_TMP_CC;
extern	int STV_TMP_CC2;


/* General Macros */

#define stv_myid( _idp ) \
	( STV_TMP_CC = pvm_setopt( PvmAutoErr, 0 ), \
		*(_idp) = pvm_mytid(), \
		pvm_setopt( PvmAutoErr, STV_TMP_CC ), \
		( *(_idp) >= 0 ) ? stvTrue : stvFalse )

#define stv_exit pvm_exit

#define stv_parent( _idp ) \
	( *(_idp) = pvm_parent(), \
		( *(_idp) >= 0 ) ? stvTrue : stvFalse )

#define stv_f_to_c_str	ftocstr


/* Host Macros */

#define stv_id_to_host( _id )  pvm_tidtohost( _id )

#define stv_add_hosts( _hosts, _nhosts, _infos ) \
	( pvm_addhosts( _hosts, _nhosts, _infos ) )

#define stv_host_status( _host ) \
	( STV_TMP_CC = pvm_mstat( _host ), \
		( STV_TMP_CC ) ? stvFalse : stvTrue )


/* Spawning Macro */

#define stv_spawn( _task, _targv, _flag, _where, _nt, _tids, _numtp ) \
	( STV_TMP_CC = \
			pvm_spawn( _task, _targv, _flag, _where, _nt, _tids ), \
		( (_numtp) != NULL ) ? ( *(_numtp) = STV_TMP_CC ) : stvTrue, \
		( STV_TMP_CC == (_nt) ) ? stvTrue : stvFalse )

#define stv_kill( _task ) \
	( STV_TMP_CC = pvm_kill( _task ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )


/* Group Macros */

#define stv_join_group( _name, _size, _instp ) \
	( *(_instp) = pvm_joingroup( _name ), \
		( *(_instp) >= 0 ) ? stvTrue : stvFalse )

#define stv_leave_group( _name, _inst, _size ) \
	( STV_TMP_CC = pvm_lvgroup( _name ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_broadcast( _bufid, _name, _tag, _ctx ) \
	( stv_set_context( _ctx, &STV_TMP_CTX ), \
		STV_TMP_CC2 = pvm_setsbuf( _bufid ), \
		STV_TMP_CC = pvm_bcast( _name, _tag ), \
		pvm_setsbuf( STV_TMP_CC2 ),\
		stv_set_context( STV_TMP_CTX, (STV_CONTEXT *) NULL ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_group_size( _name, _sizep ) \
	( *(_sizep) = pvm_gsize( _name ), \
		( *(_sizep) >= 0 ) ? stvTrue : stvFalse )

#define stv_get_id_from_instance( _group, _inst, _idp ) \
	( *(_idp) = pvm_gettid( _group, _inst ), \
		( *(_idp) >= 0 ) ? stvTrue : stvFalse )

#define stv_get_inst_from_id( _group, _id, _instp ) \
	( *(_instp) = pvm_getinst( _group, _id ), \
		( *(_instp) >= 0 ) ? stvTrue : stvFalse )

#define stv_get_group_tids( _group, _tidlistp, _ntidsp ) \
	( STV_TMP_CC = gs_get_tidlist( _group, 1, _ntidsp, &(_tidlistp), \
			stvFalse ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )


/* Name Service Macros */

#if ( PVM_MAJOR_VERSION > 3 ) || ( PVM_MINOR_VERSION >= 4 )

#define stv_name_insert( _name, _index, _tagp, _indexp ) \
	( STV_TMP_CC = pvm_setsbuf( pvm_mkbuf( PvmDataDefault ) ), \
		pvm_pkint( _tagp, 1, 1 ), \
		*(_indexp) = pvm_putinfo( _name, pvm_getsbuf(), \
				PvmMboxDefault ), \
		pvm_setsbuf( STV_TMP_CC ), \
		( *(_indexp) >= 0 ) ? stvTrue : stvFalse )

#define stv_name_lookup( _name, _index, _tagp, _indexp ) \
	( *(_indexp) = pvm_recvinfo( _name, _index, PvmMboxDefault ), \
		( *(_indexp) >= 0 ) ? \
			( STV_TMP_CC = pvm_upkint( _tagp, 1, 1 ), \
				( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse ) \
			: stvFalse )

#define stv_name_insert_ctx( _name, _index, _ctxp, _indexp ) \
	( STV_TMP_CC = pvm_setsbuf( pvm_mkbuf( PvmDataDefault ) ), \
		pvm_pkint( _ctxp, 1, 1 ), \
		*(_indexp) = pvm_putinfo( _name, pvm_getsbuf(), \
				PvmMboxPersistent | PvmMboxDefault ), \
		pvm_setsbuf( STV_TMP_CC ), \
		( *(_indexp) >= 0 ) ? stvTrue : stvFalse )

#define stv_name_lookup_ctx( _name, _index, _ctxp, _indexp ) \
	( *(_indexp) = pvm_recvinfo( _name, _index, PvmMboxDefault ), \
		( *(_indexp) >= 0 ) ? \
			( STV_TMP_CC = pvm_upkint( _ctxp, 1, 1 ), \
				( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse ) \
			: stvFalse )

#define stv_name_delete(_name, _index, _indexp ) \
	( *(_indexp) = pvm_delinfo( _name, _index, PvmMboxDefault ), \
		( *(_indexp) >= 0 ) ? stvTrue : stvFalse )

#else

#define stv_name_insert( _name, _index, _tagp, _indexp ) \
	( *(_indexp) = pvm_insert( _name, _index, *(_tagp) ), \
		( *(_indexp) >= 0 ) ? stvTrue : stvFalse )

#define stv_name_lookup( _name, _index, _tagp, _indexp ) \
	( *(_indexp) = pvm_lookup( _name, _index, _tagp ), \
		( *(_indexp) >= 0 ) ? stvTrue : stvFalse )

#define stv_name_insert_ctx( _name, _index, _ctxp, _indexp ) \
	( *(_indexp) = pvm_insert( _name, _index, *(_ctxp) ), \
		( *(_indexp) >= 0 ) ? stvTrue : stvFalse )

#define stv_name_lookup_ctx( _name, _index, _ctxp, _indexp ) \
	( *(_indexp) = pvm_lookup( _name, _index, _ctxp ), \
		( *(_indexp) >= 0 ) ? stvTrue : stvFalse )

#define stv_name_delete(_name, _index, _indexp ) \
	( *(_indexp) = pvm_delete( _name, _index ), \
		( *(_indexp) >= 0 ) ? stvTrue : stvFalse )

#endif


/* Auto-Tracing Macro */

#if ( PVM_MAJOR_VERSION > 3 ) || ( PVM_MINOR_VERSION >= 4 )

#define stv_mp_trace	{ }

#else

#define stv_mp_trace \
	{ \
		Pvmtmask trace_mask; \
\
		int xpvm_tid; \
		int inst; \
\
		/* Join to XPVM if present */ \
\
		if ( (xpvm_tid = pvm_gettid( "xpvm", 0 )) > 0 ) \
		{ \
			/* Set Self Trace Destinations & Message Codes */ \
			pvm_setopt( PvmSelfTraceCode, 666 ); \
			pvm_setopt( PvmSelfTraceTid, xpvm_tid ); \
\
			/* Set Future Children's Trace Dests & Codes */ \
			pvm_setopt( PvmTraceCode, 666 ); \
			pvm_setopt( PvmTraceTid, xpvm_tid ); \
\
			/* Generate Default Trace Mask */ \
			TEV_INIT_MASK( trace_mask ); \
			TEV_SET_MASK( trace_mask, TEV_MCAST0 ); \
			TEV_SET_MASK( trace_mask, TEV_SEND0 ); \
			TEV_SET_MASK( trace_mask, TEV_RECV0 ); \
			TEV_SET_MASK( trace_mask, TEV_NRECV0 ); \
\
			/* Set Self Trace Mask */ \
			pvm_settmask( PvmTaskSelf, trace_mask ); \
\
			/* Set Future Children's Trace Mask */ \
			pvm_settmask( PvmTaskChild, trace_mask ); \
		} \
\
		else \
			printf( "XPVM not running, no tracing\n" ); \
	} \
\

#endif


/* Message Context Macros */

#if ( PVM_MAJOR_VERSION > 3 ) || ( PVM_MINOR_VERSION >= 4 )

#define stv_create_context( _ctxp ) \
	( *(_ctxp) = pvm_newcontext(), \
		( *(_ctxp) >= 0 ) ? stvTrue : stvFalse )

#define stv_free_context( _ctxp ) \
	( STV_TMP_CC = pvm_freecontext( *(_ctxp) ), \
		*(_ctxp) = (STV_CONTEXT) -1, \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_set_context( _newctx, _oldctxp ) \
	( STV_TMP_CTX = pvm_setcontext( _newctx ), \
		( (_oldctxp) != NULL ) \
			? ( *(_oldctxp) = STV_TMP_CTX ) : stvTrue, \
		( STV_TMP_CTX >= 0 ) ? stvTrue : stvFalse )

#define stv_get_context( _ctxp ) \
	( *(_ctxp) = pvm_getcontext(), \
		( *(_ctxp) >= 0 ) ? stvTrue : stvFalse )

#else

#define stv_create_context( _ctxp ) \
	( *(_ctxp) = (STV_CONTEXT) stvTrue, stvTrue )

#define stv_free_context( _ctxp ) \
	( *(_ctxp) = (STV_CONTEXT) -1, stvTrue )

#define stv_set_context( _newctx, _oldctxp ) \
	( ( (_oldctxp) != NULL ) ? \
			( *(_oldctxp) = (STV_CONTEXT) stvTrue ) : stvTrue, \
		stvTrue )

#define stv_get_context( _ctxp ) \
	( *(_ctxp) = (STV_CONTEXT) stvTrue, stvTrue )

#endif


/* Message Buffer Macros */

#define stv_init_msg_buf( _enc, _bufp ) \
	( STV_TMP_CC = pvm_setsbuf(0), \
		*(_bufp) = pvm_initsend( _enc ), \
		pvm_setsbuf(STV_TMP_CC), \
		( *(_bufp) > 0 ) ? stvTrue : stvFalse )

#define stv_msg_buf_info( _buf, _nbytesp, _tagp, _srcp ) \
	( STV_TMP_CC = pvm_bufinfo( _buf, _nbytesp, _tagp, _srcp ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_current_sbuf(_bufp) \
	( *(_bufp) = pvm_getsbuf())

#define stv_set_sbuf(bufp) \
	(pvm_setsbuf(bufp))

#define stv_current_rbuf(_bufp) \
	( *(_bufp) = pvm_getrbuf())

#define stv_set_rbuf(bufp) \
	(pvm_setrbuf(bufp))

#define stv_free_msg_buf( _buf ) \
	( STV_TMP_CC = pvm_freebuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )


/* Packing & Unpacking Macros */


/** PACKers **/

#define stv_pack_byte( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkbyte( _datap, _count, _stride ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_int( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkint( _datap, _count, _stride ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_float( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkfloat( _datap, _count, _stride ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_double( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkdouble( _datap, _count, _stride ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_complex( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkcplx( _datap, _count, _stride ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_dcomplex( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkdcplx( _datap, _count, _stride ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_short( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkshort( _datap, _count, _stride ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_ushort( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkushort( _datap, _count, _stride ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_uint( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkuint( _datap, _count, _stride ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_long( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pklong( _datap, _count, _stride ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_ulong( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkulong( _datap, _count, _stride ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_str( _buf, _str ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkstr( _str ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_pack_context( _buf, _ctxp ) \
	( _buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_pkint( _ctxp, 1, 1 ), \
		_buf = pvm_setsbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )


/****** UN-PACKers *******/

#define stv_unpack_byte( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkbyte( _datap, _count, _stride ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_int( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkint( _datap, _count, _stride ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_float( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkfloat( _datap, _count, _stride ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_double( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkdouble( _datap, _count, _stride ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_complex( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkcplx( _datap, _count, _stride ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_dcomplex( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkdcplx( _datap, _count, _stride ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_short( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkshort( _datap, _count, _stride ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_ushort( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkushort( _datap, _count, _stride ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_uint( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkuint( _datap, _count, _stride ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_long( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upklong( _datap, _count, _stride ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_ulong( _buf, _datap, _count, _stride ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkulong( _datap, _count, _stride ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_str( _buf, _str ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkstr( _str ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_unpack_context( _buf, _ctxp ) \
	( _buf = pvm_setrbuf( _buf ), \
		STV_TMP_CC = pvm_upkint( _ctxp, 1, 1 ), \
		_buf = pvm_setrbuf( _buf ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )


/* Notify Macro */

#define stv_notify_event( _evnt, _tag, _ctx, _ntids, _tidp ) \
      ( stv_set_context( _ctx, &STV_TMP_CTX ), \
		STV_TMP_CC = pvm_notify( _evnt, _tag, _ntids, _tidp ), \
		stv_set_context( STV_TMP_CTX, (STV_CONTEXT *) NULL ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse)


/* Message Send & Receive Macros */

#define stv_send_msg( _buf, _dst, _tag, _ctx ) \
	( stv_set_context( _ctx, &STV_TMP_CTX ), \
		_buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_send( _dst, _tag ), \
		_buf = pvm_setsbuf( _buf ), \
		stv_set_context( STV_TMP_CTX, (STV_CONTEXT *) NULL ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_mcast_msg( _buf, _dstlist, _numdst, _tag, _ctx ) \
	( stv_set_context( _ctx, &STV_TMP_CTX ), \
		_buf = pvm_setsbuf( _buf ), \
		STV_TMP_CC = pvm_mcast( _dstlist, _numdst, _tag ), \
		_buf = pvm_setsbuf( _buf ), \
		stv_set_context( STV_TMP_CTX, (STV_CONTEXT *) NULL ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

#define stv_recv_msg( _src, _tag, _ctx, _bufp ) \
	( stv_set_context( _ctx, &STV_TMP_CTX ), \
		STV_TMP_CC = pvm_setrbuf(0), \
		*(_bufp) = pvm_recv( _src, _tag ), \
		pvm_setrbuf( STV_TMP_CC ), \
		stv_set_context( STV_TMP_CTX, (STV_CONTEXT *) NULL ), \
		( *(_bufp) > 0 ) ? stvTrue : stvFalse )

#define stv_recv_msg_noblock( _src, _tag, _ctx, _bufp ) \
	( stv_set_context( _ctx, &STV_TMP_CTX ), \
		STV_TMP_CC = pvm_setrbuf(0), \
		*(_bufp) = pvm_nrecv( _src, _tag ), \
		pvm_setrbuf( STV_TMP_CC ), \
		stv_set_context( STV_TMP_CTX, (STV_CONTEXT *) NULL ), \
		( *(_bufp) > 0 ) ? stvTrue : stvFalse )

#define stv_recv_msg_timeout( _src, _tag, _ctx, _tmoutp, _bufp ) \
	( stv_set_context( _ctx, &STV_TMP_CTX ), \
		STV_TMP_CC = pvm_setrbuf(0), \
		*(_bufp) = pvm_trecv( _src, _tag, _tmoutp ), \
		pvm_setrbuf( STV_TMP_CC ), \
		stv_set_context( STV_TMP_CTX, (STV_CONTEXT *) NULL ), \
		( *(_bufp) > 0 ) ? stvTrue : stvFalse )


/* Message Error Macro */

#define stv_mp_err( _text ) \
	( STV_TMP_CC = pvm_perror( _text ), \
		( STV_TMP_CC >= 0 ) ? stvTrue : stvFalse )

/* STV_MP_PVM */
#endif


#ifndef STV_MP_NAME
#define STV_MP_NAME "Unknown"
#endif


/* _STV_MP_H_ */
#endif

