
/* $Id: stv_pack.c,v 3.35 2000/06/12 16:42:26 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <math.h>


/* STV Header Files */

#include "stv.h"
#include "stvmac.h"
#include "stv_glob.h"


/* Packing Routines */


/* Pack Up Local SubRegion Elements */

int
stv_pack_SubRegion( V, FG, tag, ctx )
STV_FE_VIEWERPTR V;
STV_FIELD_GROUP_PTR FG;
int tag;
STV_CONTEXT ctx;
{
	STV_PARAM_MAP_PTR PM;

	STV_PACK_DIM PD;

	STV_FIELDPTR F;

	STV_PARAMPTR P;

	int nparams;
	int nfields;
	int nparts;
	int i, j;
	int sbuf;
	int type;
	int tmp;
	int cc;

	/* Pack Region Message */

	stv_init_msg_buf( stvDataDefault, &sbuf );

	tmp = stvMsgOp( stvRegion );
	stv_pack_int( sbuf, &tmp, 1, 1 );

	/* Pack Field Group ID */

	stv_pack_int( sbuf, &(FG->fgid), 1, 1 );

	/* Pack Viewer ID */

	stv_pack_int( sbuf, &(V->vid), 1, 1 );

	/* Pack Region ID */

	stv_pack_int( sbuf, &(FG->vRegionId), 1, 1 );

	/* Pack Lock Status */

	stv_pack_int( sbuf, &STV_APP_LOCK, 1, 1 );

	/* Pack Current Timestamp */

	stv_pack_int( sbuf, &STV_TIMESTAMP, 1, 1 );

	/* Pack SubRegion Marker */

	tmp = stvMarkerSubRegion;
	stv_pack_int( sbuf, &tmp, 1, 1 );

	/* Count Standard & Particle Fields */

	nfields = 0;

	nparts = 0;

	for ( i=0 ; i < FG->nfields ; i++ )
	{
		F = FG->field_list[i];

		if ( F != NULL && F->data_ready )
		{
			if ( F->decomp != NULL )
				nfields++;

			else if ( F->particle != NULL )
				nparts++;
		}
	}

	/* Pack All Standard Fields */

	stv_pack_int( sbuf, &nfields, 1, 1 );

	for ( i=0 ; i < FG->nfields ; i++ )
	{
		F = FG->field_list[i];

		/* Check for Existing Field */

		if ( F != NULL && F->data_ready )
		{
			/* Pack Decomposition Field Only */

			if ( F->decomp != NULL )
			{
				/* Pack Field Marker */

				tmp = stvMarkerField;
				stv_pack_int( sbuf, &tmp, 1, 1 );

				/* Pack Viewer Field ID */

				stv_pack_int( sbuf, &(FG->field_ids[i]), 1, 1 );

				/* Pack Data Type */

				stv_pack_int( sbuf, &(FG->field_types[i]), 1, 1 );

				/* Pack Field Data Storage Info */

				stv_pack_int( sbuf, &(F->storage_order), 1, 1 );

				/* Pack Current Field Time */

				stv_pack_int( sbuf, &(F->time_base), 1, 1 );
				stv_pack_int( sbuf, &(F->time_units), 1, 1 );

				/* Pack Array Dimension */

				stv_pack_int( sbuf, &(F->decomp->dataDim), 1, 1 );

				/* Get Indices for SubRegion - Local & Global */

				PD = stv_get_SubRegion_indices( F,
					FG->visRegion, FG->visCell );

				/* Pack Data */

				if ( PD != NULL )
				{
					if ( STV_OPT_VERBOSE & STV_DEBUG_PACK )
					{
						stv_dump_pack_dims( PD,
							F->storage_order,
							FG->field_storage_orders[i] );
					}

					stv_pack_elems_SubRegion( sbuf, PD, F,
						FG->field_types[i] );

					stv_free_pack_dim_list( &PD );
				}

				/* Pack End of Field Marker */

				tmp = stvMarkerFieldEnd;
				stv_pack_int( sbuf, &tmp, 1, 1 );
			}
		}
	}

	/* Pack Particle Fields */

	if ( nparts )
		stv_pack_SubRegion_Particles( sbuf, V, FG, nparts );

	else
	{
		tmp = 0;
		stv_pack_int( sbuf, &tmp, 1, 1 );
	}

	/* Pack End of SubRegion Marker */

	tmp = stvMarkerSubRegionEnd;
	stv_pack_int( sbuf, &tmp, 1, 1 );

	/* Pack Parameter Values Marker */

	tmp = stvMarkerParamValues;
	stv_pack_int( sbuf, &tmp, 1, 1 );

	/* Pack Number of Parameter Values */

	PM = V->param_map->next;  /* skip first dummy struct - for Philbo */

	nparams = 0;

	while ( PM != NULL )
	{
		nparams++;

		PM = PM->next;
	}

	stv_pack_int( sbuf, &nparams, 1, 1 );

	/* Pack Parameter Values */

	PM = V->param_map->next;  /* skip first dummy struct - for Philbo */

	while ( PM != NULL )
	{
		stv_pack_int( sbuf, &(PM->remoteID), 1, 1 );

		P = PM->local;

		stv_pack_int( sbuf, &(P->num), 1, 1 );

		for ( i=0 ; i < P->num ; i++ )
		{
			if ( (cc = stv_pack_by_type( sbuf, P->vars[i], 1, 1,
				P->types[i] & stvTypeMask )) != stvStatusOk )
			{
				STV_ERROR( "stv_pack_SubRegion", cc,
					"Typed Pack Failed." );
				
				return( cc );
			}
		}

		PM = PM->next;
	}

	/* Pack End of Parameter Values Marker */

	tmp = stvMarkerParamValuesEnd;
	stv_pack_int( sbuf, &tmp, 1, 1 );

	/* Send Message */

	stv_send_msg( sbuf, V->vtid, tag, ctx );

	stv_free_msg_buf( sbuf );

	/* A.O.K. */

	return( stvStatusOk );
}


/* Determine Indices for SubRegion */

STV_PACK_DIM
stv_get_SubRegion_indices( F, R, C )
STV_FIELDPTR F;
STV_REGION R;
STV_CELL C;
{
	STV_PACK_DIM PDptr;
	STV_PACK_DIM PDtmp;
	STV_PACK_DIM PD;

	STV_DECOMPPTR D;

	int pindex;
	int index;

	/* Initialize Return List */

	PD = (STV_PACK_DIM) NULL;

	/* Get Decomposition */

	D = F->decomp;

	/* a la Fortran */

	if ( F->storage_order == stvColumnMajor )
	{
		pindex = 0;

		for ( index=0 ; index < D->dataDim ; index++ )
		{
			PDtmp = stv_get_SubRegion_index( F, R, C, index, &pindex );

			if ( PD == NULL )
				PDptr = PD = PDtmp;
		
			else
				PDptr = PDptr->next = PDtmp;
		}
	}

	/* a la C */

	else if ( F->storage_order == stvRowMajor )
	{
		pindex = D->prank - 1;

		for ( index=(D->dataDim-1) ; index >= 0 ; index-- )
		{
			PDtmp = stv_get_SubRegion_index( F, R, C, index, &pindex );

			if ( PD == NULL )
				PDptr = PD = PDtmp;
		
			else
				PDptr = PDptr->next = PDtmp;
		}
	}

	/* Oops, Unknown Layout */

	else
	{
		sprintf( STVERRTXT, "Unknown Data Storage Order, ID = %d",
			F->storage_order );

		STV_ERROR( "stv_get_SubRegion_indices", stvStatusBadParam,
			STVERRTXT );

		return( (STV_PACK_DIM) NULL );
	}

	/* Verify Non-Null Packing Space */

	PDptr = PD;

	while ( PDptr != NULL )
	{
		if ( PDptr->pindices == NULL )
		{
			stv_free_pack_dim_list( &(PD) );

			return( (STV_PACK_DIM) NULL );
		}

		PDptr = PDptr->next;
	}

	/* Return Pack Dimension List */

	return( PD );
}


/* Determine Dimension Index for SubRegion */

STV_PACK_DIM
stv_get_SubRegion_index( F, R, C, index, pindex )
STV_FIELDPTR F;
STV_REGION R;
STV_CELL C;
int index;
int *pindex;
{
	STV_PACK_INDEX PXptr;
	STV_PACK_INDEX PX;

	STV_PACK_DIM PD;

	STV_DECOMPPTR D;

	int gnext, lnext;
	int cyclesize;
	int min, max;
	int arraymax;
	int gindex;
	int offset;
	int tmpmax;
	int first;
	int bogus;
	int delta;
	int cell;
	int q, k;
	int b;

	/* Get Decomposition */

	D = F->decomp;

	/* Create New Pack Dimension */

	PD = stv_create_pack_dim();

	STV_NO_MEM_TYPE_return( PD, "stv_get_SubRegion_index",
		STV_PACK_DIM );

	PD->type = D->axisType[ index ];

	PD->dim = index;

	/* Create Initial Pack Index */

	PX = PD->pindices = stv_create_pack_index();

	STV_NO_MEM_TYPE_return( PX, "stv_get_SubRegion_index",
		STV_PACK_DIM );

	/* Snag Min & Max for this Dimension */

	max = STV_REGION_MAX( R, index );
	min = STV_REGION_MIN( R, index );

	cell = STV_CELL_OF_DIM( C, index );

	/* Set Starts, Strides & Max based on Axis Type */

	switch ( D->axisType[ index ] )
	{
		case stvBlock:
		{
			/* Array Max */

			arraymax = D->axisInfo[ index ];

			/* Calculate Block Index */

			gindex = ( F->paddr[ *pindex ] * D->axisInfo[ index ] )
				+ D->glb[ index ];

			if ( max >= gindex + D->axisInfo[ index ] )
				PX->max = arraymax;

			else
				PX->max = max - gindex + 1;

			PX->gstart = STV_START_BLOCK_INDEX( gindex, min, cell );

			while ( PX->gstart < min
				&& PX->gstart - gindex + cell < arraymax )
			{
				PX->gstart += cell;
			}

			bogus = stvTrue;

			if ( PX->gstart >= min && PX->gstart <= max )
			{
				PX->gstride = cell;

				PX->lstart = PX->gstart - gindex;
				PX->lstride = cell;

				if ( PX->lstart <= arraymax )
					bogus = stvFalse;
			}

			if ( bogus )
				stv_free_pack_index( &(PD->pindices) );

			if ( F->storage_order == stvColumnMajor )
				(*pindex)++;
			
			else if ( F->storage_order == stvRowMajor )
				(*pindex)--;

			break;
		}

		case stvCyclic:
		{
			/* Snag Array Max Index */

			arraymax = stv_get_proc_dim_size( stvCyclic,
				D->axisInfo[ index ],
				D->axisLBounds[ index ], D->axisUBounds[ index ],
				D->nbounds[ index ],
				D->gub[ index ] - D->glb[ index ] + 1,
				F->paddr[ *pindex ], D->pshape[ *pindex ] );

			if ( arraymax < 0 )
			{
				STV_ERROR( "stv_get_SubRegion_index", arraymax,
					"stv_get_proc_dim_size() failed." );
				
				return( (STV_PACK_DIM) NULL );
			}

			/* Calculate Cyclic Indices for Each Sub-Cycle */

			cyclesize = D->axisInfo[ index ];

			PXptr = (STV_PACK_INDEX) NULL;

			for ( q=0 ; q < cyclesize ; q++ )
			{
				if ( q )
				{
					if ( PD->pindices == NULL )
						PX = PD->pindices = stv_create_pack_index();

					else
					{
						PXptr = PX;

						PX = PX->next = stv_create_pack_index();
					}

					STV_NO_MEM_TYPE_return( PX,
						"stv_get_SubRegion_index", STV_PACK_DIM );
				}

				/* Find Starting Index */

				k = ( ( min - D->glb[ index ] )
					/ ( cyclesize * D->pshape[ *pindex ] ) ) - 1;

				do
				{
					k++;

					PX->gstart = STV_GLOBAL_CYCLE_INDEX( q, k,
						cyclesize, D->glb[ index ],
						F->paddr[ *pindex ], D->pshape[ *pindex ] );
				}
				while ( PX->gstart < min
					|| ( !STV_DIVISIBLE( PX->gstart - min, cell )
						&& PX->gstart <= max ) );

				/* Verify Start Within Region */

				bogus = stvTrue;

				if ( PX->gstart <= max )
				{
					PX->lstart = STV_LOCAL_CYCLE_INDEX( q, k,
						cyclesize );

					/* Verify Start Within Local Portion */

					if ( PX->lstart < arraymax )
					{
						bogus = stvFalse;

						/* Find Next Index */

						do
						{
							k++;

							gnext = STV_GLOBAL_CYCLE_INDEX( q, k,
								cyclesize, D->glb[ index ],
								F->paddr[ *pindex ],
								D->pshape[ *pindex ] );
						}
						while ( !STV_DIVISIBLE( gnext - min, cell )
							&& gnext <= max );

						/* Calculate Stride */

						if ( gnext <= max )
						{
							PX->gstride = gnext - PX->gstart;

							lnext = STV_LOCAL_CYCLE_INDEX( q, k,
								cyclesize );
						
							PX->lstride = lnext - PX->lstart;

							/* Find Max Index */

							delta = PX->gstride
								/ ( cyclesize * D->pshape[ *pindex ] );

							do
							{
								k += delta;

								gnext = STV_GLOBAL_CYCLE_INDEX( q, k,
									cyclesize, D->glb[ index ],
									F->paddr[ *pindex ],
									D->pshape[ *pindex ] );
							}
							while ( gnext <= max );

							PX->max = STV_LOCAL_CYCLE_INDEX( q, k,
								cyclesize );
						}

						else
						{
							PX->gstride = max - min + 1;

							PX->lstride = arraymax;

							PX->max = arraymax;
						}
					}
				}

				if ( bogus )
				{
					if ( PXptr != NULL )
					{
						stv_free_pack_index( &(PXptr->next) );

						PX = PXptr;
					}

					else
						stv_free_pack_index( &(PD->pindices) );
				}
			}

			if ( bogus && PXptr == NULL && PD->pindices != NULL )
				stv_free_pack_index( &(PD->pindices) );

			if ( F->storage_order == stvColumnMajor )
				(*pindex)++;
			
			else if ( F->storage_order == stvRowMajor )
				(*pindex)--;

			break;
		}

		case stvExplicit:
		{
			PXptr = (STV_PACK_INDEX) NULL;

			offset = 0;

			for ( b=0 ; b < D->nbounds[ index ] ; b++ )
			{
				if ( b )
				{
					if ( PD->pindices == NULL )
						PX = PD->pindices = stv_create_pack_index();

					else
					{
						PXptr = PX;

						PX = PX->next = stv_create_pack_index();
					}

					STV_NO_MEM_TYPE_return( PX,
						"stv_get_SubRegion_index", STV_PACK_DIM );
				}

				/* Calculate Explicit Index */

				first = MAX( min, D->axisLBounds[ index ][ b ] );

				PX->gstart = STV_START_BLOCK_INDEX( first, min, cell );
				PX->gstride = cell;

				PX->lstart = offset
						+ ( PX->gstart - D->axisLBounds[ index ][ b ] );
				PX->lstride = cell;

				tmpmax = MIN( max, D->axisUBounds[ index ][ b ] );

				PX->max = offset
						+ ( tmpmax - D->axisLBounds[ index ][ b ] + 1 );

				/* Check for Bogosity */

				bogus = stvTrue;

				if ( PX->lstart < PX->max )
					bogus = stvFalse;

				if ( bogus )
				{
					if ( PXptr != NULL )
					{
						stv_free_pack_index( &(PXptr->next) );

						PX = PXptr;
					}

					else
						stv_free_pack_index( &(PD->pindices) );
				}

				offset += D->axisUBounds[ index ][ b ]
						- D->axisLBounds[ index ][ b ] + 1;
			}

			if ( bogus && PXptr == NULL && PD->pindices != NULL )
				stv_free_pack_index( &(PD->pindices) );

			if ( F->storage_order == stvColumnMajor )
				(*pindex)++;
		
			else if ( F->storage_order == stvRowMajor )
				(*pindex)--;

			break;
		}

		case stvCollapse:
		{
			/* Calculate Collapse Index */

			PX->gstart = min;
			PX->gstride = cell;

			PX->lstart = PX->gstart - D->glb[ index ];
			PX->lstride = cell;

			PX->max = max - D->glb[ index ] + 1;

			break;
		}
	}

	return( PD );
}


/* Dump Bounds of Local Task's Portion of Global Field Array */

int
stv_dumpLocalFieldBounds( id, verbose )
int id;
int verbose;
{
	STV_PARTICLEPTR PT;

	STV_DECOMPPTR D;

	STV_FIELDPTR F;

	int cyclesize;
	int pindex;
	int gstart;
	int gend;
	int b;
	int k;
	int i;

	/* Lookup Field */

	if ( (F = (STV_FIELDPTR) stv_findField( &STVFIELDS, id ))
		== (STV_FIELDPTR) NULL )
	{
		sprintf( STVERRTXT, "Unknown Field ID %d", id );

		STV_ERROR( "stv_dumpLocalFieldBounds", stvStatusBadParam,
			STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Check for Particle Field */

	if ( F->particle != NULL )
	{
		PT = F->particle;

		/* Dump Global Array Bounds */

		if ( verbose )
		{
			printf(
			"\nGlobal Array Bounds for Particle Field \"%s\":\n\n\t",
				F->name );
		}

		else
			printf( "Global Array \"%s\":  ", F->name );

		for ( i=0 ; i < PT->dataDim ; i++ )
			printf( "[ %d - %d ]", PT->glb[i], PT->gub[i] );

		if ( verbose )
			printf( "\n\n" );
		
		else
			printf( "\n" );
	}

	/* Regular Decomposition */

	else if ( F->decomp != NULL )
	{
		D = F->decomp;

		/* Dump Global Array Bounds */

		if ( verbose )
		{
			printf( "\nGlobal Array Bounds for Field \"%s\":\n\n\t",
				F->name );
		}

		else
			printf( "Global Array \"%s\":  ", F->name );

		for ( i=0 ; i < D->dataDim ; i++ )
			printf( "[ %d - %d ]", D->glb[i], D->gub[i] );
		
		if ( verbose )
			printf( "\n\nLogical array of processors is:\n\n\t" );
		
		else
			printf( "\nProcessor Array:  " );

		for ( i=0 ; i < D->prank ; i++ )
		{
			printf( "%d%s", D->pshape[i],
				( i < ( D->prank - 1 ) ? "  X  " : "" ) );
		}

		if ( verbose )
			printf( "\n\nLocal processor address is:\n\n\t{ ");

		else
			printf( ",  Local Address:  { " );

		for ( i=0 ; i < D->prank ; i++ )
		{
			printf( "%d%s", F->paddr[i],
				( i < ( D->prank - 1 ) ? ", " : " }" ) );
		}

		/* matching } */

		/* Dump Local Decomp Bounds */

		if ( verbose )
		{
			printf(
				"\n\nLocal Portion of Global Space for Field \"%s\":",
				F->name );
		}

		pindex = 0;

		for ( i=0 ; i < D->dataDim ; i++ )
		{
			if ( verbose )
				printf( "\n\n\tBounds for Axis #%d:\n\n\t\t", i + 1 );
			
			else
				printf( "\nAxis #%d:  ", i + 1 );

			switch ( D->axisType[i] )
			{
				case stvBlock:
				{
					/* Calculate Block Index */

					gstart = ( F->paddr[ pindex ] * D->axisInfo[i] )
						+ D->glb[i];

					gend = gstart + D->axisInfo[i] - 1;

					printf( "[ %d - %d ]", gstart, gend );

					pindex++;

					break;
				}

				case stvCyclic:
				{
					/* Calculate Cyclic Indices for Each Sub-Cycle */

					cyclesize = D->axisInfo[i];

					/* Find Starting Index */

					k = 0;

					do
					{
						gstart = STV_GLOBAL_CYCLE_INDEX( 0, k,
							cyclesize, D->glb[i],
							F->paddr[ pindex ], D->pshape[ pindex ] );

						gend = gstart + cyclesize - 1;

						gend = ( gend > D->gub[i] ) ?
							D->gub[i] : gend;

						if ( gstart <= D->gub[i] )
						{
							if ( gend != gstart )
								printf( "[ %d - %d ]", gstart, gend );

							else
								printf( "[ %d ]", gstart );
						}

						k++;
					}
					while ( gstart <= D->gub[i] );

					pindex++;

					break;
				}

				case stvExplicit:
				{
					/* Calculate Explicit Index */

					for ( b=0 ; b < D->nbounds[i] ; b++ )
					{
						gstart = D->axisLBounds[i][b];

						gend = D->axisUBounds[i][b];

						printf( "[ %d - %d ]", gstart, gend );
					}

					pindex++;
			
					break;
				}

				case stvCollapse:
				{
					/* Calculate Collapse Index */

					gstart = D->glb[i];

					gend = D->gub[i];

					printf( "[ %d - %d ]", gstart, gend );

					break;
				}

				default:
				{
					sprintf( STVERRTXT,
						"Unknown Axis Type %d for Axis #%d",
						D->axisType[i], i + 1 );

					STV_ERROR( "stv_dumpLocalFieldBounds",
						stvStatusBadParam, STVERRTXT );

					return( stvStatusBadParam );
				}
			}
		}

		if ( verbose )
			printf( "\n\n" );

		else
			printf( "\n" );
	}

	return( stvStatusOk );
}


/* Pack Data Elements for SubRegion */

int
stv_pack_elems_SubRegion( sbuf, PD, F, type )
int sbuf;
STV_PACK_DIM PD;
STV_FIELDPTR F;
int type;
{
	STV_PACK_INDEX PX;

	STV_PACK_DIM PDptr;

	STV_VALUE ptr;

	int offset;
	int num;
	int tmp;
	int cc;

	/* Initialize Indices */

	stv_init_pack_indices( PD );

	/* Iterate & Pack Stuff */

	do
	{
		/* Pack End of Data Marker */

		tmp = stvMarkerFieldData;
		stv_pack_int( sbuf, &tmp, 1, 1 );

		/* Pack Offset Index */

		stv_pack_offset_index( sbuf, PD );

		/* Get Iterative Index Info */

		PX = PD->current;

		/* Calculate Number of Elements to Pack */

		num = (int) ceil( (double) ( PX->max - PX->lstart )
			/ (double) PX->lstride );

		/* Pack Unpack Info */

		if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
		{
			printf( "%s: packing gstart=%d gstride=%d num=%d\n",
				"stv_pack_elems_SubRegion()",
				PX->gstart, PX->gstride, num );
			fflush( stdout );
		}

		stv_pack_int( sbuf, &(PX->gstart), 1, 1 );
		stv_pack_int( sbuf, &(PX->gstride), 1, 1 );
		stv_pack_int( sbuf, &(num), 1, 1 );

		/* Pack Data */

		offset = stv_get_pack_offset( PD, F );

		ptr = stv_get_typed_ptr_offset( F->var, offset + PX->lstart,
			F->type );

		if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
		{
			printf( "%s: data ptr = %lx (F->var = %lx).\n",
				"stv_pack_elems_SubRegion()",
				(long) ptr, (long) F->var );
			fflush( stdout );
		}

		if ( ptr == NULL )
		{
			STV_ERROR( "stv_pack_elems_SubRegion", stvStatusBadParam,
				"stv_get_typed_ptr_offset() failed." );
			
			return( stvStatusBadParam );
		}

		if ( F->type != type )
		{
			if ( (cc = stv_pack_converted_type( sbuf, ptr, num,
				PX->lstride, F->type, type )) != stvStatusOk )
			{
				STV_ERROR( "stv_pack_elems_SubRegion", cc,
					"Converted Type Pack Failed." );
				
				return( cc );
			}
		}

		else
		{
			if ( (cc = stv_pack_by_type( sbuf, ptr, num, PX->lstride,
				type )) != stvStatusOk )
			{
				STV_ERROR( "stv_pack_elems_SubRegion", cc,
					"Typed Pack Failed." );
				
				return( cc );
			}
		}

		/* Pack End of Data Marker */

		tmp = stvMarkerFieldDataEnd;
		stv_pack_int( sbuf, &tmp, 1, 1 );
	}
	while ( stv_incr_pack_indices( PD ) );

	/* A.O.K. */

	return( stvStatusOk );
}


/* Initialize Pack Indices */

void
stv_init_pack_indices( PD )
STV_PACK_DIM PD;
{
	STV_PACK_DIM ptr;

	ptr = PD;

	while ( ptr != NULL )
	{
		ptr->current = ptr->pindices;

		ptr->gindex = ptr->current->gstart;
		ptr->lindex = ptr->current->lstart;

		ptr = ptr->next;
	}
}


/* Pack N - 1 Offset Index Values */

int
stv_pack_offset_index( sbuf, PD )
int sbuf;
STV_PACK_DIM PD;
{
	STV_PACK_DIM ptr;

	if ( PD == NULL )
	{
		STV_ERROR( "stv_pack_offset_index", stvStatusBadParam,
			"Null Pack Dimension Array" );

		return( stvStatusBadParam );
	}

	ptr = PD;

	/* Skip First Iterative Index */

	ptr = ptr->next;

	/* Pack In Rest */

	if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
		printf( "stv_pack_offset_index():  [X]" );

	while ( ptr != NULL )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
			printf( "[%d]", ptr->gindex );

		stv_pack_int( sbuf, &(ptr->gindex), 1, 1 );

		ptr = ptr->next;
	}

	if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
	{
		printf( "\n" );
		fflush( stdout );
	}

	return( stvStatusOk );
}


/* Get Pack Offset Into Data Array */

int
stv_get_pack_offset( PD, F )
STV_PACK_DIM PD;
STV_FIELDPTR F;
{
	STV_PACK_DIM ptr;

	int offset;
	int size;

	if ( PD == NULL )
		return( 0 );

	/* Init Offset & Size */

	if ( F->arrayOffset != NULL )
		offset = F->arrayOffset[ PD->dim ];
	
	else
		offset = 0;

	size = F->arrayDecl[ PD->dim ];

	/* Skip First Iterative Index */

	ptr = PD->next;

	/* Add Up Rest */

	while ( ptr != NULL )
	{
		if ( F->arrayOffset != NULL )
		{
			offset +=
					( ptr->lindex + F->arrayOffset[ ptr->dim ] ) * size;
		}

		else
			offset += ptr->lindex * size;

		size *= F->arrayDecl[ ptr->dim ];

		ptr = ptr->next;
	}

	return( offset );
}


/* Increment Packing Indices */

int
stv_incr_pack_indices( PD )
STV_PACK_DIM PD;
{
	STV_PACK_DIM PDptr;

	int first;
	int done;

	if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
		printf( "stv_incr_pack_indices():  " );

	if ( PD == NULL )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
		{
			printf( "Null Pack Dim.\n" );
			fflush( stdout );
		}
		
		return( stvFalse );
	}

	/* Skip Over First Index Unless Cyclic Steps Needed */

	first = 1;

	if ( PD->current->next == NULL && PD->current == PD->pindices )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
			printf( "SKIP,  " );

		PDptr = PD->next;

		first = 0;
	}
	
	/* Else ACTUALLY SET the PDptr to the First Index...  D-Oh! */

	else
		PDptr = PD;

	/* Verify there's anything to increment... */

	if ( PDptr == NULL )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
		{
			printf( "Null Pack Dim.\n" );
			fflush( stdout );
		}
		
		return( stvFalse );
	}

	/* Excrement Away! */

	do
	{
		if ( PDptr->current == NULL )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
			{
				printf( "Current Pack Index Null.\n" );
				fflush( stdout );
			}
		
			return( stvFalse );
		}

		if ( !first )
		{
			PDptr->gindex += PDptr->current->gstride;
			PDptr->lindex += PDptr->current->lstride;

			if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
			{
				printf( "Incr gindex[%d] -> %d, lindex[%d] -> %d,  ",
						PDptr->dim, PDptr->gindex,
						PDptr->dim, PDptr->lindex );
			}
		}

		/* Check for Bump */

		done = stvTrue;

		if ( PDptr->lindex >= PDptr->current->max || first )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
				printf( "BUMP(max=%d),  ", PDptr->current->max );

			/* Check for Cyclic/Explicit Multi-Index Bump */

			if ( PDptr->current->next != NULL )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
					printf( "Next,  " );

				PDptr->current = PDptr->current->next;
			}

			else
			{
				/* Check for Cyclic/Explicit Multi-Index Reset */

				if ( PDptr->current != PDptr->pindices )
				{
					if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
						printf( "Reset,  " );

					PDptr->current = PDptr->pindices;
				}

				done = stvFalse;
			}

			/* Reset Index */

			PDptr->gindex = PDptr->current->gstart;
			PDptr->lindex = PDptr->current->lstart;

			if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
			{
				printf( "gindex[%d]=%d, lindex[%d]=%d,  ",
						PDptr->dim, PDptr->gindex,
						PDptr->dim, PDptr->lindex );
			}

			if ( !done )
			{
				PDptr = PDptr->next;

				if ( !PDptr )
				{
					if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
						printf( "End of Pack Dims.\n" );

					return( stvFalse );
				}
			}
		}

		first = 0;
	}
	while ( !done );

	if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
	{
		printf( "Done.\n" );
		fflush( stdout );
	}

	return( stvTrue );
}


/* Convert the Data to Requested Type & Pack */

int
stv_pack_converted_type( sbuf, data, num, stride, data_type, pack_type )
int sbuf;
STV_VALUE data;
int num;
int stride;
int data_type;
int pack_type;
{
	STV_VALUE copy_data;
	STV_VALUE src;
	STV_VALUE dst;

	int cc;
	int i;

	/* Allocate Copy Storage */

	copy_data = (STV_VALUE) MALLOC( (STV_SIZE_T) num
		* STV_SIZEOF( pack_type ) );
	
	if ( !copy_data )
	{
		STV_ERROR( "stv_pack_converted_type", stvStatusNoMem,
			"Couldn't Allocate Conversion Buffer" );
		
		return( stvStatusNoMem );
	}

	/* Copy Data & Cast to Pack Type */

	src = data;
	dst = copy_data;

	for ( i=0 ; i < num ; i++ )
	{
		/* Convert One Data Element */

		if ( (cc = stv_copy_converted_data( src, data_type,
			dst, pack_type )) != stvStatusOk )
		{
			STV_ERROR( "stv_pack_converted_type", cc,
				"Converted Type Copy Failed." );

			FREE( copy_data );

			return( cc );
		}

		/* Advance Ptrs */

		src = stv_get_typed_ptr_offset( src, stride, data_type );

		dst = stv_get_typed_ptr_offset( dst, 1, pack_type );

		if ( src == NULL || dst == NULL )
		{
			STV_ERROR( "stv_pack_converted_type", stvStatusBadParam,
				"stv_get_typed_ptr_offset() failed." );

			FREE( copy_data );

			return( stvStatusBadParam );
		}
	}

	/* Pack Copied Converted Data, No Longer Strided */

	if ( (cc = stv_pack_by_type( sbuf, copy_data, num, 1, pack_type ))
		!= stvStatusOk )
	{
		STV_ERROR( "stv_pack_converted_type", cc,
			"Typed Pack Failed." );

		FREE( copy_data );

		return( cc );
	}

	/* Free Copy */

	FREE( copy_data );

	/* A.O.K. */

	return( stvStatusOk );
}


/* Dump the Pack Dimension Indices Info */

void
stv_dump_pack_dims( PD, field_storage_order, view_storage_order )
STV_PACK_DIM PD;
int field_storage_order;
int view_storage_order;
{
	STV_PACK_INDEX PX;

	STV_PACK_DIM PDptr;

	int i;

	printf( "Desired Data Storage Order is %s:\n",
		stv_storage_order_str( view_storage_order ) );

	printf( "Field Data Storage Order is %s:\n",
		stv_storage_order_str( field_storage_order ) );

	PDptr = PD;

	i = 0;

	while ( PDptr != NULL )
	{
		PX = PDptr->pindices;

		printf(
		"\tDIM #%d: gstart=%d gstride=%d lstart=%d lstride=%d max=%d\n",
			i++, PX->gstart, PX->gstride, PX->lstart, PX->lstride,
			PX->max );

		PX = PX->next;

		while ( PX != NULL )
		{
			printf(
			"\t\tgstart=%d gstride=%d lstart=%d lstride=%d max=%d\n",
				PX->gstart, PX->gstride, PX->lstart, PX->lstride,
				PX->max );

			PX = PX->next;
		}
		
		PDptr = PDptr->next;
	}
}


/* Unpack Data Elements from SubRegion */

int
stv_unpack_elems_SubRegion( rbuf, VFG, V )
int rbuf;
STV_VIEW_FIELD_GROUP VFG;
STV_VIEWER V;
{
	STV_VIEW_FIELD VF;

	STV_VIEW_PARAM VP;

	STV_DECOMPPTR D;

	STV_PARAMPTR P;

	STV_REGION FR;

	STV_VALUE ptr;

	double dummy[10];

	int indices[STV_MAX_DIM];

	int field_storage_order;
	int start, stride, num;
	int storage_stride;
	int nparams;
	int nfields;
	int nparts;
	int first;
	int index;
	int i, j;
	int step;
	int type;
	int dim;
	int tmp;
	int cc;
	int id;

	/* Unpack SubRegion Marker */

	stv_unpack_int( rbuf, &tmp, 1, 1 );

	if ( tmp != stvMarkerSubRegion )
	{
		sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
			"stvMarkerSubRegion", stvMarkerSubRegion, tmp );

		STV_ERROR( "stv_unpack_elems_SubRegion",
			stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Unpack # of Standard Fields */

	stv_unpack_int( rbuf, &nfields, 1, 1 );

	if ( STV_OPT_VERBOSE & STV_DEBUG_PACK )
	{
		printf( "%s:  Unpacking %d Standard Fields\n",
			"stv_unpack_elems_SubRegion", nfields );
		fflush( stdout );
	}

	/* Unpack Fields */

	for ( i=0 ; i < nfields ; i++ )
	{
		/* Unpack Field Marker */

		stv_unpack_int( rbuf, &tmp, 1, 1 );

		if ( tmp != stvMarkerField )
		{
			sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
				"stvMarkerField", stvMarkerField, tmp );

			STV_ERROR( "stv_unpack_elems_SubRegion",
				stvStatusBadParam, STVERRTXT );

			return( stvStatusBadParam );
		}

		/* Unpack Field ID */

		stv_unpack_int( rbuf, &id, 1, 1 );

		if ( STV_OPT_VERBOSE & STV_DEBUG_PACK )
		{
			printf( "Unpacked Field ID = %d\n", id );
			fflush( stdout );
		}

		/* Get Field */

		VF = stv_get_view_field_from_group_id( VFG, id, &index );

		if ( VF == NULL )
		{
			sprintf( STVERRTXT, "Unknown Field ID %d", id );

			STV_ERROR( "stv_unpack_elems_SubRegion",
				stvStatusBadParam, STVERRTXT );

			return( stvStatusBadParam );
		}

		if ( STV_OPT_VERBOSE & STV_DEBUG_PACK )
		{
			printf( "Field is \"%s\"\n", VF->field->name );
			fflush( stdout );
		}

		/* Verify Specific Region for This Field */

		stv_copy_region( FR, VFG->visregion );

		stv_check_region( VF->field, FR, VFG->viscell );

		/* Get Decomposition */
	
		D = VF->field->decomp;
	
		/* Unpack Data Type Check */

		stv_unpack_int( rbuf, &type, 1, 1 );

		if ( VFG->view_types[index] != type )
		{
			sprintf( STVERRTXT, "Incorrect Unpack Type %s (%d) for %s",
				stv_data_type_str( type ), type, VF->field->name );

			STV_ERROR( "stv_unpack_elems_SubRegion",
				stvStatusBadParam, STVERRTXT );

			return( stvStatusBadParam );
		}

		/* Unpack Data Storage Order */

		stv_unpack_int( rbuf, &field_storage_order, 1, 1 );

		/* Unpack Current Field Time */

		stv_unpack_int( rbuf, &(VF->field->time_base), 1, 1 );
		stv_unpack_int( rbuf, &(VF->field->time_units), 1, 1 );

		if ( STV_OPT_VERBOSE & STV_DEBUG_TIME )
		{
			printf( "Field %s:  time = ", VF->field->name );
			stv_dump_base_time_str( stdout,
					VF->field->time_base, VF->field->time_units,
					stvTrue );
			fflush( stdout );
		}

		/* Unpack Data Dimension Check */

		stv_unpack_int( rbuf, &dim, 1, 1 );

		if ( dim != D->dataDim )
		{
			sprintf( STVERRTXT, "Dim Mismatch %d/%d for %s",
				dim, D->dataDim, VF->field->name );

			STV_ERROR( "stv_unpack_elems_SubRegion",
				stvStatusBadParam, STVERRTXT );

			return( stvStatusBadParam );
		}

		/* Unpack Field Data Marker */

		stv_unpack_int( rbuf, &tmp, 1, 1 );

		if ( tmp == stvMarkerFieldEnd )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_PACK )
			{
				printf( "\tNo Data\n" );
				fflush( stdout );
			}
		}

		/* Unpack Field Data */

		while ( tmp == stvMarkerFieldData )
		{
			/* Get Rest of Offset Indices */

			for ( j=1 ; j < D->dataDim ; j++ )
				stv_unpack_int( rbuf, &(indices[j]), 1, 1 );

			if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
			{
				printf( "Unpacking Index [X]" );
				for ( j=1 ; j < D->dataDim ; j++ )
					printf( "[%d]", indices[j] );
				printf( ": " );
			}

			/* Get Unpack Info */

			stv_unpack_int( rbuf, &(start), 1, 1 );
			stv_unpack_int( rbuf, &(stride), 1, 1 );
			stv_unpack_int( rbuf, &(num), 1, 1 );

			if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
			{
				printf( "gstart=%d gstride=%d num=%d\n",
					start, stride, num );
				fflush( stdout );
			}

			/* Unpack Data */

			indices[0] = start;

			first = stv_get_unpack_offset( indices,
				VFG->field_arrayDecls[index],
				D, FR, VFG->viscell,
				VFG->view_storage_orders[index],
				field_storage_order );

			storage_stride = stv_get_storage_stride(
				VFG->field_arrayDecls[index], D,
				VFG->view_storage_orders[index],
				field_storage_order );

			stride *= storage_stride;

			if ( STV_OPT_FULL_UNPACK )
				step = 1;

			else
			{
				if ( field_storage_order == stvColumnMajor )
					step = STV_CELL_OF_DIM( VFG->viscell, 0 );

				else if ( field_storage_order == stvRowMajor )
				{
					step = STV_CELL_OF_DIM( VFG->viscell,
						D->dataDim - 1 );
				}

				else
				{
					sprintf( STVERRTXT, "Unknown View Storage Order %d",
						field_storage_order );

					STV_ERROR( "stv_unpack_elems_SubRegion",
						stvStatusBadParam, STVERRTXT );

					return( stvStatusBadParam );
				}
			}

			stride /= step;

			/* Get Unpack Ptr */

			ptr = stv_get_typed_ptr_offset(
				VFG->field_values[index], first, type );

			if ( ptr == NULL )
			{
				STV_ERROR( "stv_unpack_elems_SubRegion",
					stvStatusBadParam,
					"stv_get_typed_ptr_offset() failed." );
				
				return( stvStatusBadParam );
			}

			if ( STV_OPT_VERBOSE & STV_DEBUG_PACK_INDEX )
			{
				printf( "Data ptr = %lx (field_values[%d] = %lx).\n",
					(long) ptr, index,
					(long) VFG->field_values[index] );
				fflush( stdout );
			}

			/* Unpack Typed Data */

			if ( (cc = stv_unpack_by_type( rbuf, ptr, num,
				stride, type )) != stvStatusOk )
			{
				STV_ERROR( "stv_unpack_elems_SubRegion", cc,
					"Typed Unpack Failed." );
				
				return( cc );
			}

			/* Unpack End of Data Marker */

			stv_unpack_int( rbuf, &tmp, 1, 1 );

			if ( tmp != stvMarkerFieldDataEnd )
			{
				sprintf( STVERRTXT,
					"Parse Error: Expected %s = %d, Got %d",
					"stvMarkerFieldDataEnd", stvMarkerFieldDataEnd,
					tmp );

				STV_ERROR( "stv_unpack_elems_SubRegion",
					stvStatusBadParam, STVERRTXT );

				return( stvStatusBadParam );
			}

			/* Unpack Next Marker */

			stv_unpack_int( rbuf, &tmp, 1, 1 );
		}

		if ( tmp != stvMarkerFieldEnd )
		{
			sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
				"stvMarkerFieldEnd", stvMarkerFieldEnd, tmp );

			STV_ERROR( "stv_unpack_elems_SubRegion",
				stvStatusBadParam, STVERRTXT );

			return( stvStatusBadParam );
		}
	}

	/* Unpack # of Particle Fields */

	stv_unpack_int( rbuf, &nparts, 1, 1 );

	if ( STV_OPT_VERBOSE & STV_DEBUG_PACK )
	{
		printf( "%s:  Unpacking %d Particle Fields\n",
			"stv_unpack_elems_SubRegion", nparts );
		fflush( stdout );
	}

	/* Unpack Particle Fields */

	for ( i=0 ; i < nparts ; i++ )
	{
		if ( (cc = stv_unpack_elems_SubRegion_Particles( rbuf,
			VFG )) != stvStatusOk )
		{
			return( cc );
		}
	}

	/* Unpack End of SubRegion Marker */

	stv_unpack_int( rbuf, &tmp, 1, 1 );

	if ( tmp != stvMarkerSubRegionEnd )
	{
		sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
			"stvMarkerSubRegionEnd", stvMarkerSubRegionEnd, tmp );

		STV_ERROR( "stv_unpack_elems_SubRegion",
			stvStatusBadParam, STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Check for Parameter Values Update Section */

	stv_unpack_int( rbuf, &tmp, 1, 1 );

	if ( tmp == stvMarkerParamValues )
	{
		/* Unpack # of Parameter Value Updates */

		stv_unpack_int( rbuf, &nparams, 1, 1 );

		if ( STV_OPT_VERBOSE & STV_DEBUG_PACK )
		{
			printf( "%s:  Unpacking %d Parameter Value Updates\n",
				"stv_unpack_elems_SubRegion", nparams );
			fflush( stdout );
		}

		/* Unpack Parameter Value Updates */

		for ( i=0 ; i < nparams ; i++ )
		{
			stv_unpack_int( rbuf, &id, 1, 1 );

			VP = stv_get_view_param_id( id, V->vparams, V->nvparams );

			P = VP->param;

			if ( VP != NULL && P != NULL )
			{
				stv_unpack_int( rbuf, &num, 1, 1 );

				if ( num != P->num )
				{
					sprintf( STVERRTXT,
						"%s:  \"%s\" Num Mismatch (%d != %d) - Ignore.",
						"Parameter Value Update", P->name,
						num, P->num );

					STV_ERROR( "stv_unpack_elems_SubRegion",
						stvStatusBadParam, STVERRTXT );

					return( stvStatusBadParam );
				}

				for ( j=0 ; j < num ; j++ )
				{
					ptr = ( VP->token == stvSteerToken )
						? ( (STV_VALUE) dummy ) : ( P->vars[j] );

					if ( (cc = stv_unpack_by_type( rbuf, ptr, 1, 1,
						P->types[j] & stvTypeMask )) != stvStatusOk )
					{
						sprintf( STVERRTXT,
						"%s: \"%s\" Typed Unpack Failed (%d of %d).",
							"Parameter Value Update", P->name,
							j, num );

						STV_ERROR( "stv_unpack_elems_SubRegion", cc,
							STVERRTXT );

						return( cc );
					}
				}
			}
		}

		/* Unpack End of SubRegion Marker */

		stv_unpack_int( rbuf, &tmp, 1, 1 );

		if ( tmp != stvMarkerParamValuesEnd )
		{
			sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
				"stvMarkerParamValuesEnd", stvMarkerParamValuesEnd,
				tmp );

			STV_ERROR( "stv_unpack_elems_SubRegion",
				stvStatusBadParam, STVERRTXT );

			return( stvStatusBadParam );
		}
	}

	return( stvStatusOk );
}


/* Generate Unpack Offset */

int
stv_get_unpack_offset( indices, arrayDecl, D, R, C,
	view_storage_order, field_storage_order )
int *indices;
int *arrayDecl;
STV_DECOMPPTR D;
STV_REGION R;
STV_CELL C;
int view_storage_order;
int field_storage_order;
{
	int b, e, s;
	int offset;
	int index;
	int size;
	int step;
	int min;
	int i;

	if ( view_storage_order == stvColumnMajor )
	{
		b = 0;
		e = D->dataDim;
		s = 1;
	}

	else if ( view_storage_order == stvRowMajor )
	{
		b = D->dataDim - 1;
		e = -1;
		s = -1;
	}

	offset = 0;

	size = 1;

	for ( i=b ; i != e ; i += s )
	{
		if ( STV_OPT_FULL_UNPACK )
		{
			min = D->glb[i];

			step = 1;
		}

		else
		{
			min = STV_REGION_MIN( R, i );

			step = STV_CELL_OF_DIM( C, i );
		}

		if ( field_storage_order == stvRowMajor )
			index = D->dataDim - i - 1;
		
		else
			index = i;

		offset += ( ( indices[ index ] - min ) / step ) * size;

		size *= arrayDecl[i];
	}

	return( offset );
}


/* Get Cross-Storage-Order Striding Factor */

int
stv_get_storage_stride( arrayDecl, D,
	view_storage_order, field_storage_order )
int *arrayDecl;
STV_DECOMPPTR D;
int view_storage_order;
int field_storage_order;
{
	int stride;
	int i;

	stride = 1;

	if ( view_storage_order != field_storage_order )
	{
		if ( view_storage_order == stvColumnMajor )
		{
			for ( i=0 ; i < D->dataDim - 1 ; i++ )
				stride *= arrayDecl[i];
		}

		else if ( view_storage_order == stvRowMajor )
		{
			for ( i=(D->dataDim - 1) ; i > 0 ; i-- )
				stride *= arrayDecl[i];
		}
	}

	return( stride );
}

